/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file algorithms.h
 *Here are defined various reachability algorithms...
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */
#ifndef ALGORITHMS_H
#define ALGORITHMS_H

#include "configuration.h"
#include "transition_operation.h"

#define ALG_DEBUG

#ifdef ALG_DEBUG
#include <iostream>
#endif


/**
 *class for backward reachability over upward closed sets of configurations.
 *
 */
class SBR
{
  public:
  SBR(const UCSetConf* _i, const UCSetConf* _f, const Pre* _p):Initial(_i), Final(_f), Spre(_p){}//! user must provide an initial and a final sets of configuration and a pre operator over T with return_type an upward closed set of configurations. 

  public:
  bool run()
  {

    UCSetConf V0,V1,V2; 
    V1.insert(Final->begin() , Final->end());//W=F and V=empty
    V2=V1;

    UCSetConf Wt, V012; //temporary set Wt
    V012=V1;



#ifdef ALG_DEBUG
    int iter_number=0;
    std::cout << "iteration : " << iter_number << " ";
    std::cout << *Final << std::endl;
    //std::cout<< " final size: " <<Final->size() << std::endl;
#endif
    
    bool flag=false;
    UCSetConf::const_iterator c=Initial->begin(),c_end=Initial->end();
    for(;c!=c_end;++c)
      {
	if (V012.includes(*c)) {flag=true;}
      }

    //while((!(V2.empty())) && (!(V012.includes(*Initial)))) // not(V<=W) or not(W<=I)
    while((!(V2.empty())) && (!flag))
      {
	V0.insert(V1.begin(), V1.end());
	V1=V2;
	V2.clear();

	UCSetConf::const_iterator p=V1.begin() , p_end=V1.end();
	for( ; p!=p_end ; p++)
	  {
#ifdef ALG_DEBUG
	    //std::cout << "# predecessors of " << **p << std::endl ;//<< Wt << std::endl; 
#endif	    
	    Wt=Spre->operator()(*p);

#ifdef ALG_DEBUG
	    //	    std::cout << "# predecessors of " << **p << std::endl << Wt << std::endl; 
#endif	    
	      
	    UCSetConf::const_iterator w=Wt.begin() , w_end=Wt.end();
	    for(; w!=w_end; w++)
	      {
		//		std::cout << "v012" << V012 << std::endl;
		

		if(!(V012.includes(*w)))
		  {
		    V2.insert(*w);
		    V012.insert(*w);
		  }
	      }

	    Wt.clear();
	    
	    UCSetConf::const_iterator c=Initial->begin(),c_end=Initial->end();
	    for(;c!=c_end;++c)
	      {
		if (V012.includes(*c)) {flag=true;}
	      }
	  }//W=W+pre(W)

#ifdef ALG_DEBUG
    ++iter_number;
    std::cout <<"## iteration : " << iter_number << "####################################" << std::endl;
    //std::cout << V2 << std::endl  << std::endl ;
    std::cout << "number of constraints: " << V012.size() << std::endl ;
#endif
      }

    std::cout << "backward reachable from F in the approximated transition system: "<< std::endl;
    std::cout <<  V012 << std::endl ;
    
    return flag;
  }
  
  private:
  const UCSetConf* Initial;
  const UCSetConf* Final;
  const Pre* Spre;
};

/**
 *class for forward simulation starting from an upward closed sets of configurations.
 *
 */
class SFO
{
  public:
  SFO(const UCSetConf* _i, const Post* _p):Initial(_i),  Spost(_p){}

  public:
  bool run(unsigned int number_step) //!the algorithm takes the number of steps as input
  {

    UCSetConf V,W; 
    W.insert(Initial->begin() , Initial->end());//W=F and V=empty
    UCSetConf Wt; //temporary set Wt
    unsigned int iter_number=0;

#ifdef ALG_DEBUG
  std::cout << "iteration : " << iter_number << " ";
  std::cout << *Initial << std::endl;
#endif
    
    while(iter_number <= number_step) // number of iteration less than the number of steps
      {
	V=W;//V=W
	UCSetConf::const_iterator p=V.begin() , p_end=V.end();
	for( ; p!=p_end ; p++)
	  {
	    Wt=Spost->operator()(*p);
	    W.insert(Wt.begin() , Wt.end());
	    Wt.clear();
	  }//W=W+post(W)

#ifdef ALG_DEBUG
    ++iter_number;
    std::cout << "iteration : " << iter_number << " ";
    std::cout << W << std::endl;
#endif

      }

    return true;
  }
 
  private:
  const UCSetConf* Initial;
  const Post* Spost;
};

#endif
