/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file condition.h
 *ADT for conditions.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef CONDITION_H
#define CONDITION_H

#include "process.h"
#include "predicate.h"
#include "starpredicate.h"
#include "starquantifier.h"
#include "printable.h"
#include "exconst.h"

#include <assert.h>

namespace guard
{
  /**
   *Typedef for the parameter type of condition. Each condition gets 4 
   *parameters: 3 of them are of type iterator over a word of process symbols, 
   *and the last is of type symbol. 
   */
  using namespace process;
  using namespace predicate;
  using namespace starpredicate;
  using namespace starquantifier;

  typedef PWord::const_iterator Const_In;

  /**
   *class condition.
   */
  class Condition:public virtual Printable
  {
  public:
    class constraint_generator
    {
    public:
      typedef GExConst value_type;
      typedef Ref<constraint_generator> ref_type;

    public:
      virtual bool end() const=0;
      virtual constraint_generator& operator++()=0;
      virtual constraint_generator& operator=(const Condition::constraint_generator& other)=0;
      virtual value_type operator*() const=0;
      virtual ~constraint_generator() {}
    };
    
  public:
    typedef constraint_generator iterator_type;

  public:
    virtual bool operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const=0;
    virtual constraint_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const=0;
    virtual ~Condition() {};
  };

  /**
   *class local condition.
   */
  class Local_Condition:public virtual Condition
  {
  public:
    virtual ~Local_Condition() {};
  };
  
  /**
   *class global condition.
   */
  class Global_Condition:public virtual Condition
  {
  public:
    virtual ~Global_Condition() {};
  };
  
  /**
   *class  TLocal_Condition.
   */
  class TLocal_Condition:public virtual Local_Condition
  {
  public:
    TLocal_Condition(const Predicate* _p):p(_p) {}
      //     ~TLocal_Condition(){delete(p);}

      bool operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	return p->operator()(*pos);
      }

      void print(std::ostream& out) const {out << *p;}
  
  public:
      class constraint_generator:public virtual Condition::constraint_generator
      {
      public:
	typedef Condition::constraint_generator::value_type value_type;

      private:
	typedef StarPredicate<Process>::constraint_iterator_ref sub_type;
      
      public:
	constraint_generator() {}
	constraint_generator(const StarPredicate<Process>* _p, const PWord* _w, const Const_In _pos, const Global_Symbol* _gc):w(_w),gc(_gc),pos(_pos),next_pos(_pos),s(sub_type(&_p->begin(*pos))),atEnd(false)
	  {
	    if (!_p->operator()(*pos)) {atEnd=true;}
	    if (pos!=w->end()) {++next_pos;}
	  }
	  
	  bool end() const
	  {
	    return atEnd;
	  }
	  
	  constraint_generator& operator++()
	    {
	      if (atEnd) {return *this;}
	      ++(*s);
	      if ((*s).end()) {atEnd=true;}
	      return *this;
	    }
	  
	  constraint_generator& operator=(const Condition::constraint_generator& other)
	    {
	      const constraint_generator* r=dynamic_cast<const constraint_generator*>(&other);
	      w=r->w;pos=r->pos;next_pos=r->next_pos;s=r->s;atEnd=r->atEnd;
	      return *this;
	    }
	  
	  
	  value_type operator*() const
	  {
	    PWord temp_w;
	    temp_w.insert(temp_w.begin(),w->begin(),pos);
	    temp_w.push_back(**s);
	    temp_w.insert(temp_w.end(),next_pos,w->end());
	    value_type result(temp_w,*gc);result.word.adjust(w->begin(),pos,w->end());
	    return result;
	  }
      private:
	  const PWord* w;
	  const Global_Symbol* gc;
	  Const_In pos;
	  Const_In next_pos;
	  sub_type s;
	  bool atEnd;
      };

  public:
      constraint_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
	{
	  const StarPredicate<Process>* r=dynamic_cast<const StarPredicate<Process>*>(p);
	  assert(r);
	  return *new constraint_generator(r,w,pos,gc);
	}

  private:
      const Predicate* p;    
  };
  
  /**
   *class TGlobal_Condition with quantifier.
   */
  class TGlobal_Condition_Quantified:public virtual Global_Condition
  {
  public:
    TGlobal_Condition_Quantified(const Quantifier<Const_In>* _q):q(_q) {}
      ~TGlobal_Condition_Quantified(){delete(q);}

      bool operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	return q->operator()(w->begin(), w->end(), pos);
      }

      void print(std::ostream& out) const {out << *q;}
  
  public:
      class constraint_generator:public virtual Condition::constraint_generator
      {
      public:
	typedef Condition::constraint_generator::value_type value_type;

      private:
	typedef StarQuantifier<Process,Const_In>::constraint_iterator_ref sub_type;
      
      public:
	constraint_generator() {}
	constraint_generator(const StarQuantifier<Process,Const_In>* _p, const PWord* _w, const Const_In pos, const Global_Symbol* _gc):gs(_gc),s(sub_type(&_p->begin(_w->begin(),_w->end(),pos))),atEnd(false) 
	  {
	    if (!_p->operator()(_w->begin(),_w->end(),pos)) {atEnd=true;}
	  }
	  
	  bool end() const
	  {
	    return atEnd;
	  }
	  
	  constraint_generator& operator++()
	    {
	      if (atEnd) {return *this;}
	      ++(*s);
	      if ((*s).end())  {atEnd=true;}
	      return *this;
	    }
	  
	  constraint_generator& operator=(const Condition::constraint_generator& other)
	    {
	      const constraint_generator* r=dynamic_cast<const constraint_generator*>(&other);
	      gs=r->gs;s=r->s;atEnd=r->atEnd;
	      return *this;
	    }
	  
	  value_type operator*() const
	  {
	    value_type result;
	    result.word=**s;
	    result.gs=*gs;
	    return result;
	  }

      private:
	  const Global_Symbol* gs;
	  sub_type s;
	  bool atEnd;
      };

  public:
      constraint_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
	{
	  const StarQuantifier<Process,Const_In>* r=dynamic_cast<const StarQuantifier<Process,Const_In>*>(q);
	  assert(r);
	  return *new constraint_generator(r,w,pos,gc);
	}   

  private:
      const Quantifier<Const_In>* q;
  };
  
  class TGlobal_Condition:public virtual Global_Condition
  {
  public:
    TGlobal_Condition(const Predicate* _p):p(_p) {}
      //    ~TGlobal_Condition(){delete(p);}
    bool operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
    {
      return (*p).operator()(*gc);
    }
    
    void print(std::ostream& out) const {out << *p;}
public:
      class constraint_generator:public virtual Condition::constraint_generator
      {
      public:
	typedef Condition::constraint_generator::value_type value_type;

      private:
	typedef StarPredicate<Global_Symbol>::constraint_iterator_ref sub_type;
      
      public:
	constraint_generator() {}
	constraint_generator(const StarPredicate<Global_Symbol>* _p, const PWord* _w, const Const_In _pos, const Global_Symbol* _gc):w(_w),pos(_pos),s(sub_type(&_p->begin(*_gc))),atEnd(false) 
	  {
	    if (!_p->operator()(*_gc)) {atEnd=true;}
	  }
	  
	  bool end() const
	  {
	    return atEnd;
	  }
	  
	  constraint_generator& operator++()
	    {
	      if (atEnd) {return *this;}
	      ++(*s);
	      if ((*s).end()) {atEnd=true;}
	      return *this;
	    }
	  
	  constraint_generator& operator=(const Condition::constraint_generator& other)
	    {
	      const constraint_generator* r=dynamic_cast<const constraint_generator*>(&other);
	      w=r->w;s=r->s;pos=r->pos;atEnd=r->atEnd;
	      return *this;
	    }
	  
	  value_type operator*() const
	  {
	    value_type result(*w,**s);
	    result.word.adjust(w->begin(),pos,w->end());
	    return result;
	  }

      private:
	  const PWord* w;
	  Const_In pos;
	  sub_type s;
	  bool atEnd;
      };

    constraint_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
	{
	  const StarPredicate<Global_Symbol>* r=dynamic_cast<const StarPredicate<Global_Symbol>*>(p);
	  assert(r);
	  return *new constraint_generator(r,w,pos,gc);
	}

  private:
    const Predicate* p;
  };

  class TNew_Condition:public virtual Local_Condition
  {
  public:
    TNew_Condition(){}
    bool operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
    {
      return true;
    }

      void print(std::ostream& out) const {out << " new ";}
  
  public:
      class constraint_generator:public virtual Condition::constraint_generator
      {
      public:
	typedef Condition::constraint_generator::value_type value_type;

      public:
	constraint_generator() {}
	constraint_generator(const PWord* _w, const Const_In _pos, const Global_Symbol* _gc):w(_w),gc(_gc),pos(_pos),atEnd(false)
	  {
	    if(pos!=w->end()) {
	      next_pos=++pos;
	    }
	    if(pos!=w->begin()) {prev_pos=--pos;}
	    //	    atEnd=true;
	  }
	  
	  bool end() const
	  {
	    return atEnd;
	  }
	  
	  constraint_generator& operator++()
	    {
	      atEnd=true;
	      return *this;
	    }
	  
	  constraint_generator& operator=(const Condition::constraint_generator& other)
	    {
	      const constraint_generator* r=dynamic_cast<const constraint_generator*>(&other);
	      w=r->w; pos=r->pos; next_pos=r->next_pos; prev_pos=r->prev_pos; atEnd=r->atEnd;
	      return *this;
	    }
	  
	  
	  value_type operator*() const
	  {
	    PWord temp_w;
	    temp_w.insert(temp_w.begin(),w->begin(),prev_pos);
	    temp_w.insert(temp_w.end(),next_pos,w->end());
	    value_type result(temp_w,*gc);
	    return result;
	  }
      private:
	  const PWord* w;
	  const Global_Symbol* gc;
	  Const_In pos;
	  Const_In next_pos;
	  Const_In prev_pos;
	  bool atEnd;
      };

  public:
      constraint_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
	{
	  return *new constraint_generator(w,pos,gc);
	}

  };



}
  
#endif

  


