/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file configuration.h
 *ADT for processconfiguration data structure in general.
 *Our configuration are going to be words associated with a global condition.
 *In our case the word is going to be a list-based word of symbols to capture the 
 *arrangment and the status of our process, and the global condition is a symbol 
 *itself which points out the status of the global variables.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */
#ifndef CONFIGURATION_H
#define CONFIGURATION_H
#include <iostream>

#include "symbol.h"
#include "word.h"
#include "ref.h"
#include "printable.h"
#include "minset.h"

/**
 *class configuration.
 */
class Configuration:public Printable
{
 public:
  Configuration() {} 
  
  Configuration(const Word* _w, const Symbol* _gc): processes(WordRef(_w)),global_condition(SymRef(_gc)) {}
    
    Configuration(const Configuration& other):processes(other.processes),global_condition(other.global_condition) {}
      
      ~Configuration() {}
      
 public:
      
      const WordRef processes;
      const SymRef global_condition;
      
      const Word* get_word() const {return &(*processes);}
      const Symbol* get_global() const {return &(*global_condition);}

 public:
      
      void print(std::ostream& out) const 
      {
	//	out<<"[" << *global_condition << "" << "\n" << *processes <<"]";
	out<<"\n" <<"[" << *global_condition << "" << *processes <<"]";
      }
      
};

typedef Ref<Configuration> ConfRef;

/**
 *This class order on configurations is needed to handle upward closed set of conf.
 */
class ConfOrder
{

 public:
  ConfOrder() {}

  bool operator()(const Configuration& C1, const Configuration& C2) const
  {
    return (C1.global_condition == C2.global_condition) && ( C1.processes <= C2.processes);
  }
  bool operator()(const ConfRef C1, const ConfRef C2) const
  {
    return ((*C1).global_condition == (*C2).global_condition) && ( (*C1).processes <= (*C2).processes);
  }

};

/**
 *Equality
 */
class ConfEquality
{

 public:
  ConfEquality() {}

  bool operator()(const Configuration& C1, const Configuration& C2) const
  {
    return (*(C1.global_condition) == *(C2.global_condition)) && ( C1.processes == C2.processes);
  }
  bool operator()(const ConfRef C1, const ConfRef C2) const
  {
    return (*((*C1).global_condition) == *((*C2).global_condition)) && ( (*C1).processes == (*C2).processes);
  }

};

/**
 *Upward Closed set of configurations.
 */
class UCSetConf:public MinSet<ConfRef , ConfOrder>
{

 public:
  UCSetConf() {}
};

/**
 *Set of configurations.
 */
class SetConf:public MinSet<ConfRef , ConfEquality>
{
 public:
  SetConf() {}

};

#endif

