/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file operation.h
 *ADT for operations.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */
#ifndef OPERATION_H
#define OPERATION_H
#include "process.h"
#include "transform.h"
#include "rtransform.h"

namespace action
{
  /**
   *Typedef for the parameter type of condition. Each condition gets 4 
   *parameters: 3 of them are of type iterator over a word of process symbols, 
   *and the last is of type symbol. 
   */
  using namespace process;
  using namespace transform;
  using namespace rtransform;

  typedef PWord::const_iterator Const_In;
  typedef PWord::iterator In;

  /**
   *structure output for the constraint generator of condition.
   *It is needed to keep track of the current position.
   */
  class GResult
  {
  public:
    GResult() {}
    GResult(const PWord& _w,const Global_Symbol& _g):word(Result<Process>(_w)),gs(_g){}

  public:
    Result<Process> word;
    Global_Symbol gs;
  };
  
  /**
   *class condition.
   */
  class Operation
  {
  public:
    class constraint_generator
    {
      public:
      virtual bool end() const=0;
      virtual constraint_generator& operator++()=0;
      virtual constraint_generator& operator=(const Operation::constraint_generator& other)=0;
      virtual GResult operator*() const=0;
      virtual ~constraint_generator() {}
    };

  public:
    typedef Ref<constraint_generator> const_iter_ref;

  public:
    virtual bool operator()(const PWord& w, const Const_In pos, const Global_Symbol& gc) const=0;
    virtual constraint_generator& begin(const PWord& w, const Const_In pos, const Global_Symbol& gc) const=0;
    virtual ~Operation() {};
  };

  /**
   *class local condition.
   */
  class Local_Operation:public virtual Operation
  {
  public:
    virtual ~Local_Operation() {};
  };
  
  /**
   *class global condition.
   */
  class Global_Operation:public virtual Operation
  {
  public:
    virtual ~Global_Operation() {};
  };
  
  /**
   *class  TLocal_Operation.
   */
  class TLocal_Operation:public virtual Local_Operation
  {
  public:
    TLocal_Operation(const Transform<Process>* _p):p(Transform<Process>::ref_type(_p)) {}
      bool operator()(const PWord& w, const Const_In pos, const Global_Symbol& gc) const
      {
	return (*p).operator()(*pos);
      }

  public:
      class constraint_generator:public virtual Operation::constraint_generator
      {
      private:
	typedef RTransform<Process>::make_hold_ref sub_type;
      
      public:
	constraint_generator() {}
	constraint_generator(const RTransform<Process>* _p, const PWord* _w, const Const_In& _pos, const Global_Symbol* _gc):w(_w),gc(_gc),pos(_pos),next_pos(_pos),s(sub_type(&_p->begin(*pos)))
	  {
	    if (pos!=w->end()) {++next_pos;}
	  }
	  
	  bool end() const
	  {
	    return (*s).end();
	  }
	  
	  constraint_generator& operator++()
	    {
	      if (!(*s).end()) {++(*s);}
	      return *this;
	    }
	  
	  constraint_generator& operator=(const Operation::constraint_generator& other)
	    {
	      const constraint_generator* r=dynamic_cast<const constraint_generator*>(&other);
	      w=r->w;pos=r->pos;next_pos=r->next_pos;s=r->s;
	      return *this;
	    }
	  
	  
	  GResult operator*() const
	  {
	    PWord temp_w;temp_w.insert(temp_w.begin(),w->begin(),pos);temp_w.push_back(**s);temp_w.insert(temp_w.end(),next_pos,w->end());
	    GResult result(temp_w,*gc);result.word.adjust(w->begin(),w->end(),pos);
	    return result;
	  }
      private:
	  const PWord* w;
	  const Global_Symbol* gc;
	  Const_In pos;
	  Const_In next_pos;
	  sub_type s;
      };

  public:
      constraint_generator& begin(const PWord& w, const Const_In pos, const Global_Symbol& gc) const
	{
	  const RTransform<Process>* r=dynamic_cast<const RTransform<Process>*>(&(*p));
	  return *new constraint_generator(r,&w,pos,&gc);
	}

  private:
      const Transform<Process>::ref_type p;    
  };
  
  
  class TGlobal_Operation:public virtual Global_Operation
  {
  public:
    TGlobal_Operation(const Transform<Global_Symbol>* _p):p(Transform<Global_Symbol>::ref_type(_p)) {}
    bool operator()(const PWord& w, const Const_In pos, const Global_Symbol& gc) const
    {
      return (*p).operator()(gc);
    }
    
public:
      class constraint_generator:public virtual Operation::constraint_generator
      {
      private:
	typedef RTransform<Global_Symbol>::make_hold_ref sub_type;
      
      public:
	constraint_generator() {}
	constraint_generator(const RTransform<Global_Symbol>* _p, const PWord* _w, const Const_In& _pos, const Global_Symbol* _gc):w(_w),pos(_pos),s(sub_type(&_p->begin(*_gc))) {}
	  
	  bool end() const
	  {
	    return (*s).end();
	  }
	  
	  constraint_generator& operator++()
	    {
	      if (!(*s).end()) { ++(*s);}
	      return *this;
	    }
	  
	  constraint_generator& operator=(const Operation::constraint_generator& other)
	    {
	      const constraint_generator* r=dynamic_cast<const constraint_generator*>(&other);
	      w=r->w;s=r->s;pos=r->pos;
	      return *this;
	    }
	  
	  GResult operator*() const
	  {
	    GResult result(*w,**s);
	    result.word.adjust(w->begin(),w->end(),pos);
	    return result;
	  }

      private:
	  const PWord* w;
	  Const_In pos;
	  sub_type s;
      };

    constraint_generator& begin(const PWord& w, const Const_In pos, const Global_Symbol& gc) const
	{
	  const RTransform<Global_Symbol>* r=dynamic_cast<const RTransform<Global_Symbol>*>(&(*p));
	  return *new constraint_generator(r,&w,pos,&gc);
	}

  private:
    const Transform<Global_Symbol>::ref_type p;
  };

#endif
