/*  This file is part of yatesParse.

    yatesParse is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    yatesParse is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with yatesParse; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <iostream>
#include <string>
#include <vector>

/**
 * Contains the functions and objects for dealing with arguments   
 */

class ArgumentHandler {
  public :
    /**
     * Constructor, you must call this to create an object of class ArgumentHandler for your program.
     * \param tmpGenerateHelp 1 if you want --help to output the automatically generated help, 0 if you don't\n
     * e.g. ArgumentHandler program(1);\n
     * Note: program will exit if this is set to 1 and --help is called
     */
    ArgumentHandler(bool tmpGenerateHelp);	// Constructor
    /**
      * generates --help information from add_option, set_usage information, and set_program_information, then prints. Used internally only.\n
      * This is called if you enable --help by setting tmpGenerateHelp to 1, as is explained in the constructor section of these docs.\n
     */
    void print();
    /**
     * Adds a command line option to the program
     * \param tmpLongOption the 'long' version of the option (e.g. --foo)
     * \param tmpShortOption the 'short' version of the option (e.g. -f), set to NULL if you don't want a shortoption
     * \param tmpCallThis the address of the function to be called when the argument is detected, due to current limitations, this function MUST be an int. (e.g. &function_name)
     * \param tmpOptionDescription the option's description, used for --help
     * \param tmpCallOnce 0 if the argument can be called an unlimited amount of times, 1 if the argument should only be allowed to be called once
     * \param tmpValueExists 0 if the argument doesn't have a value, 1 if the argument has a value (e.g. --file stuff.txt)\n
     * Example: program.add_option("--bacon","-b",&test,"bacon is good",1,1);
     */
    void add_option(std::string tmpLongOption, std::string tmpShortOption, int (*tmpCallThis)(std::string argumentValue), std::string tmpOptionDescription, bool tmpCallOnce, bool tmpValueExists);
    /**
     * Parses the command line options and calls the necessary functions as specified with add_option\n
     * note: the program's main() function must be declared with argc as an unsigned int\n
     * e.g. int main(unsigned int argc, char* argv[]) 
     * \param argc main's argc (should be an unsigned int)
     * \param argv main's argv
     */
    void parse(unsigned int argc, char* argv[]);
    /**
     * Sets the program usage information to be displayed by --help\n
     * e.g. the $x part of "Usage: $x"
     * \param tmpUsageInformation a std::string containing the information to be displayed
     */
    void set_usage_information(std::string tmpUsageInformation);
    /** Sets the program information to be displayed by --help \n
     * Can contain program name, author, copyright, licensing information, etc. newlines are allowed.
     * \param tmpProgramInformation a std::string containing the above information\n
     */
    void set_program_information(std::string tmpProgramInformation);
    /** Can be used to set whether or not --help should be automatically created and printed via the print function after the ArgumentHandler object has been created.
     * \param tmpGenerateHelp 1 if you wish --help to be automatically created and printed if it is passed to your program, 0 if you wish to handle any --help arguments yourself
     */
    void set_generate_help(bool tmpGenerateHelp);
  private :
    unsigned int numberArgs;
    unsigned int argumentValueNumber;
    std::string usageInformation;
    std::string programInformation;
    std::vector<std::string> longOption;
    std::vector<std::string> shortOption;
    std::vector<std::string> argumentValue;
    std::vector<int (*)(std::string argumentValue)> callThis;
    std::vector<std::string> optionDescription;
    std::vector<bool> callOnce;
    std::vector<bool> beenCalled;
    std::vector<bool> valueExists;
    bool generateHelp;
    int argstrcmp(char *x, char *y); 	//so we don't need glibc's strcmp
};
