/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file predicate.h
 *ADT for predicates.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef PREDICATE_H
#define PREDICATE_H

#include "symbol.h"
#include "my_functional.h"
#include "printable.h"

namespace predicate
{
  
  /**
   *class predicate over a symbol.
   */
  class Predicate:public virtual Printable
  {
  public:
    typedef bool return_type;

  public:
    virtual bool holds_for_subset(const Symbol& s) const=0;//!holds holds_for_subset (star enabled case)
    virtual bool operator()(const Symbol& s) const=0;//!simply holds
    virtual ~Predicate() {}
  };

  /**Obvious derived predicates*/
  class True:public virtual Predicate
  {
  public:
    True() {}
    bool holds_for_subset(const Symbol& s) const {return true;}
    bool operator()(const Symbol& s) const {return true;}
    void print(std::ostream& out) const {out << 1;}
  };
  
  class False:public virtual Predicate
  {
  public:
    False() {}
    bool holds_for_subset(const Symbol& s) const {return false;}
    bool operator()(const Symbol& s) const {return false;}
    void print(std::ostream& out) const {out << 0;}
  };
  
  /**
   *class function over symbols.
   */
  class Function:public virtual Printable
  {
  public:
    typedef int return_type;

  public:
    virtual bool holds_for_subset(const Symbol& s) const=0;
    virtual int operator()(const Symbol& s) const=0;
    virtual ~Function() {}
  };

  /**
   *class composed predicate over a symbol.
   *It is parameterized by the type of the sub_predicates.
   */
  template<class Pred>
    class Composed_Predicate:public virtual Predicate
  {
  public:
    virtual bool holds_for_subset(const Symbol& s) const=0;//!holds holds_for_subset (star enabled case)
    virtual bool operator()(const Symbol& s) const=0;//!holds
    virtual const Pred* get_left() const=0; //!return left predicate
    virtual const Pred* get_right() const=0; //!return right predicate
    virtual ~Composed_Predicate() {}
  };

  /**
   *class predicate negator.
   */
  class Negator:public virtual Predicate
  {
  public:
    Negator(const Predicate* _p):p(_p) {}

      bool holds_for_subset(const Symbol& s) const
      {
	return p->holds_for_subset(s);
      }

      bool operator()(const Symbol& s) const
      {
	return !(*p)(s);
      }

     void print(std::ostream& out) const {out << "!(" << *p << ")";} 

  private:
    const Predicate* p;
  };
  
  /**
   *More complicated derived predicates ,i.e., of the form pred1 op pred2 
   *where op in {&&,!!,==,!=,<,>,<=,>=} or derived for functions.
   *This may allow ambigous instantiation: int && int
   */
  template<class Pred>
    class OpPredicate
  {
  public:
    typedef my_stl::My_Binary_Op<typename Pred::return_type> op_type;

  public:
    OpPredicate(const Pred* _p1, const Pred* _p2, const op_type* _op):p1(_p1),p2(_p2), op(_op) {}
      ~OpPredicate(){
	//delete(p1); 
	//delete(p2);
	delete(op);
      }
      const Pred* get_left() const {return p1;}
      const Pred* get_right() const {return p2;}
      const op_type* get_op() const {return op;}
      void print(std::ostream& out) const {out << "( " << *p1 << " " << *op << " " << *p2 << " )";}
      
  private:
      const Pred* p1;
      const Pred* p2;
      const op_type* op;
  };
  
  /**
   *Derived to force the user. 
   */
  class IntOpPredicate:private OpPredicate<Function>, public virtual Composed_Predicate<Function>
  {
    private:
    typedef OpPredicate<Function> base_type;
    
    public:
    IntOpPredicate(const Function* _f1, const Function* _f2,const base_type::op_type* _op):base_type(_f1,_f2,_op){}
  
    bool holds_for_subset(const Symbol& s) const
    {
      return get_left()->holds_for_subset(s) || get_right()->holds_for_subset(s);
    }

    bool operator()(const Symbol& s) const
    {
      return base_type::get_op()->operator()(get_left()->operator()(s),get_right()->operator()(s)) || holds_for_subset(s);
    }
    
    const Function* get_left() const {return base_type::get_left();}
    const Function* get_right() const {return base_type::get_right();}
    void print(std::ostream& out) const {base_type::print(out);}

  };
  
  class BinOpPredicate:private OpPredicate<Predicate>,public virtual Composed_Predicate<Predicate>
  { 
  private:
    typedef OpPredicate<Predicate> base_type;
    
  public:
    BinOpPredicate(const Predicate* _p1, const Predicate* _p2, const base_type::op_type* _op):base_type(_p1,_p2,_op) {}
      
      bool holds_for_subset(const Symbol& s) const
      {
      const my_stl::Or* r=dynamic_cast<const my_stl::Or*>(base_type::get_op());
      if (r) {return (get_left()->holds_for_subset(s) && !get_right()->operator()(s)) || (get_right()->holds_for_subset(s) && !get_left()->operator()(s)) || (get_right()->holds_for_subset(s) && get_left()->holds_for_subset(s));}
      else {return (get_left()->holds_for_subset(s) || get_right()->holds_for_subset(s)) && operator()(s);}
      }

      bool operator()(const Symbol& s) const
      {
	return base_type::get_op()->operator()(get_left()->operator()(s),get_right()->operator()(s));
      }
      
    const Predicate* get_left() const {return base_type::get_left();}
    const Predicate* get_right() const {return base_type::get_right();}
    void print(std::ostream& out) const {base_type::print(out);}
  };
  
  
  /**
   *first order quantifier.
   */
  template<class Seq>
    class Quantifier:public virtual Printable
  {
  public:
    virtual bool holds_for_subset(const Seq first, const Seq last, const Seq curr) const=0;
    virtual bool operator()(const Seq first, const Seq last, const Seq curr) const=0;
    virtual ~Quantifier() {}
  };
  
  /**
   *It exists such that a certain predicate holds.
   */
  template<class Seq,class Pred> 
    class ItExistsStrategy
  {
  public:
    ItExistsStrategy(const Pred* _p): p(_p){}
      
      bool holds_for_subset(const Seq first,const Seq last) const
      {
	Seq t=first;
	for(; t != last; t++)
	  {
	    if ((*p).operator()(*t) && !(*p).holds_for_subset(*t)) {return false;}
	  }
	return true;
      }

      bool operator()(const Seq first,const Seq last) const
      {
	Seq t=first;
	for(; t != last; t++)
	  {
	    if ((*p).operator()(*t)) {return true;}
	  }
	return true;
      }
      
      const Pred* get_pred() const {return p;}
      
  private:
      const Pred* p;
  };
  
  /**
   *It exists to the left.
   */
  template<class Seq,class Pred> 
    class ItExistsL:private ItExistsStrategy<Seq,Pred>, public virtual Quantifier<Seq>
  {
  private:
    typedef ItExistsStrategy<Seq,Pred> base_type;
    
  public:
    ItExistsL(const Pred* _p): base_type(_p) {}
      
      bool holds_for_subset(const Seq first, const Seq last, const Seq curr) const
      {
	return base_type::holds_for_subset(first,curr);
      }

      bool operator()(const Seq first, const Seq last, const Seq curr) const
      {
	return base_type::operator()(first,curr);
      }
      
      const Pred* get_pred() const {return base_type::get_pred();}
      
      void print(std::ostream& out) const {out << "ItExists p to the left such that : " << *get_pred();}
  };
  
  /**
   *It exists to the right.
   */
  template<class Seq,class Pred>
    class ItExistsR:private ItExistsStrategy<Seq,Pred>, public virtual Quantifier<Seq>
  {
 
    typedef ItExistsStrategy<Seq,Pred> base_type;
    
  public:
    ItExistsR(const Pred* _p):base_type(_p) {}
      
      bool holds_for_subset(const Seq first, const Seq last, const Seq curr) const
      {
	if (curr==last) {return true;}
	else
	  {
	    Seq t=curr;t++;
	    return base_type::holds_for_subset(t,last);
	  }
      }

      bool operator()(const Seq first, const Seq last, const Seq curr) const
      {
	if (curr==last) {return true;}
	else
	  {
	    Seq t=curr;t++;
	    return base_type::operator()(t,last);
	  }
      }
      
      const Pred* get_pred() const {return base_type::get_pred();}

      void print(std::ostream& out) const {out << "ItExists p to the right such that : " << *get_pred();}
  };
  
  /**
   *It exists through the whole range.
   */
  template<class Seq,class Pred>
    class ItExistsA:private ItExistsStrategy<Seq,Pred>, public virtual Quantifier<Seq>
  {
 
    typedef ItExistsStrategy<Seq,Pred> base_type;
    
  public:
    ItExistsA(const Pred* _p):base_type(_p) {}
      
      bool holds_for_subset(const Seq first, const Seq last, const Seq curr) const
      {
	return base_type::holds_for_subset(first,last);
      }

      bool operator()(const Seq first, const Seq last, const Seq curr) const
      {
	return base_type::operator()(first,last);
      }
      
      const Pred* get_pred() const {return base_type::get_pred();}
  
      void print(std::ostream& out) const {out << "ItExists p such that : " << *get_pred();}
  };
  
  /**
   *ForAll strategy.
   */
  template<class Seq,class Pred> 
    class ForAllStrategy
  {
  public:
    ForAllStrategy(const Pred* _p):p(_p) {}
      
      bool holds_for_subset(const Seq first, const Seq last) const
      {
	Seq t=first;
	for(; t != last; t++)
	  {
	    if ((*p).holds_for_subset(*t)) {return operator()(first,last);}
	  }
	return false;
      }

      bool operator()(const Seq first, const Seq last) const
      {
	Seq t=first;
	for(; t != last; t++)
	  {
	    if (!((*p).operator()(*t))) {return false;}
	  }
	return true;
      }
      
      const Pred* get_pred() const {return p;}
      
  private:
      const Pred* p;
  };
  
  /**
   *For all through the whole range.
   */
  template<class Seq,class Pred> 
    class ForAllA:private ForAllStrategy<Seq,Pred>, public virtual Quantifier<Seq>
  {
  private:
    typedef ForAllStrategy<Seq,Pred> base_type;

  public:
    ForAllA(const Pred* _p): base_type(_p) {}

      bool holds_for_subset(const Seq first,const Seq last,const Seq curr) const
      {
	return base_type::holds_for_subset(first,last);
      }

      bool operator()(const Seq first,const Seq last,const Seq curr) const
      {
	return base_type::operator()(first,last);
      }

      const Pred* get_pred() const {return base_type::get_pred();} 
    
      void print(std::ostream& out) const {out << "ForAll p : " << *get_pred();}
  };

  /**
   *For all to the left.
   */
  template<class Seq,class Pred> 
    class ForAllL:private ForAllStrategy<Seq,Pred>, public virtual Quantifier<Seq>
  {
  private:
    typedef ForAllStrategy<Seq,Pred> base_type;

  public:
    ForAllL(const Pred* _p): base_type(_p) {}

      bool holds_for_subset(const Seq first,const Seq last,const Seq curr) const
      {
	return base_type::holds_for_subset(first,curr);
      }

      bool operator()(const Seq first,const Seq last,const Seq curr) const
      {
	return base_type::operator()(first,curr);
      }

      const Pred* get_pred() const {return base_type::get_pred();} 

      void print(std::ostream& out) const {out << "ForAll p to the left : " << *get_pred();}
  };
  
  /**
   *For all to the right.
   */
  template<class Seq,class Pred> 
    class ForAllR:private ForAllStrategy<Seq,Pred>, public virtual Quantifier<Seq>
  {
  private:
    typedef ForAllStrategy<Seq,Pred> base_type;

  public:
    ForAllR(const Pred* _p): base_type(_p) {}

      bool holds_for_subset(const Seq first, const Seq last, const Seq curr) const
      {
	if (curr==last) {return true;}
	else
	  {
	    Seq t=curr;t++;
	    return base_type::holds_for_subset(t,last);
	  }
      }

      bool operator()(const Seq first, const Seq last, const Seq curr) const
      {
	if (curr==last) {return true;}
	else
	  {
	    Seq t=curr;t++;
	    return base_type::operator()(t,last);
	  }
      }

      const Pred* get_pred() const {return base_type::get_pred();} 

      void print(std::ostream& out) const {out << "ForAll p to the right : " << *get_pred();}
  };
 
}

#endif

