/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file process.h
 *In this file, we define all data structures needed for the complete definition of a parameterized system.
 *In particular, we define ADTs for processes, global condition, pre and post operations.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef PROCESS_H
#define PROCESS_H

#include "symbol.h"
#include "tword.h"

#include <map>
#include <string>
#include <utility>

//#define PRO_DEBUG
#ifdef PRO_DEBUG
#include <assert.h>
#endif

namespace process
{
  /**
   *Variables related type definitions.
   */
  typedef std::string Var_Id;
  
  /**
   *A variable range is a pair of integers:low limit and high limit.
   */
  typedef std::pair<int,int> Var_Range;
  
  /**
   *A variable definition is a pair of : range and bool to state wether it is
   *allowed to use the star symbol or not. In the case it is allowed the star 
   *symbol will correspond to the variable high limit + 1.
   */
  typedef std::pair<Var_Range,bool> Var_Def; 
  
  typedef std::pair<Var_Id,Var_Def> Id_Def;
  /**
   *A variable instance is a pair: name and value.
   */
  typedef std::pair<Var_Id,int> Var_Val;
  
  /**
   *class Variable table.
   *This class is used for defining the variables within a process (local variables) and the global variables.
   */
  class Variable_Table:public std::map<Var_Id,Var_Def>,public virtual Printable
    {
    private:
      typedef std::map<Var_Id,Var_Def> Container;
      
    public:
      typedef Container::iterator iterator;
      typedef Container::const_iterator const_iterator;

    public:
      Variable_Table() {}//!empty table
      
      void define_variable(const Var_Id id, int inf, int sup, bool enable_star)//!set a variable by giving the name the lower bound the upper bound and wether it can be set to star.
      {
#ifdef PRO_DEBUG
	assert(inf <= sup);
#endif
	Var_Range r(inf,sup);
	Var_Def rs(r,enable_star);
	Id_Def rsf(id,rs);
	insert(rsf);
      }
      
      bool defined(const Var_Id id) const//!check wether a variable with the given name is already in the table
      {
	return find(id)!=end();
      }

      bool in_range(const Var_Id id,int v) const//!wether a given value is within the range of a given variable id
      {
#ifdef PRO_DEBUG
	assert(find(id)!=end());
#endif 
	std::map<Var_Id,Var_Def>::const_iterator p=find(id);
	return ((v <= p->second.first.second) && (p->second.first.first<=v)) || ((p->second.second) && (v==p->second.first.second+1));
      }

      int lower_bound(const Var_Id id) const//!lower bound of the "id" variable
      {
#ifdef PRO_DEBUG
	assert(find(id)!=end());
#endif 
	return find(id)->second.first.first; 
      };
      
      int upper_bound(const Var_Id id) const//!upper bound of the "id" variable
      {
#ifdef PRO_DEBUG
	assert(find(id)!=end());
#endif 
	return find(id)->second.first.second; 
      }; 

      int star_value(const Var_Id id) const//!the possible star value for the "id" variable
      {
#ifdef PRO_DEBUG
	assert(find(id)!=end());
#endif 
	return find(id)->second.first.second+1; 
      }; 

      bool star_enabled(const Var_Id id) const//!wether "id" can be set to star
      {
#ifdef PRO_DEBUG
	assert(find(id)!=end());
#endif 
	return find(id)->second.second; 
      };

      void print(std::ostream& out) const
      {
	for(const_iterator vit=begin(); vit != end(); vit++)
	  {
	    out << vit->first << ":" << vit->second.second << " ";
	    //	    out << vit->second.second << " ";
	  }
	
      };

    };
  
  /**
   *class process: it is map of variable instances. It maps the id to the 
   *values. In our case, this ADT is used also for representing constraints. For example when
   *one of the variables is set to the star value then this corresponds to the set of all possible 
   *instantions of processes where in each one of them that variable has a possible value (not the star value).
   *
   *A process is also a symbol. Therefore, we must implement the equality operator on processes.
   *That is irrelevent for the configuration order.
   */
  class Process:public Symbol, public std::map<Var_Id,int>     
    {
    private:
      typedef std::map<Var_Id,int> Container;
      
    public:
      typedef Container::iterator iterator;
      typedef Container::const_iterator const_iterator;
      
    public:
      Process() {}//!empty process constructor
      Process(const Variable_Table* _vt): def_table(_vt)//!constructor from a given definition table 
	{
	  for(std::map<Var_Id, Var_Def>::const_iterator vit=def_table->begin(); vit != def_table->end(); vit++)
	    {
	      /**if the star is enabled then the corresponding variable is set to the star value otherwise it is set to the lower bound**/
	      if(def_table->star_enabled(vit->first))
		{set_variable(vit->first, def_table->star_value(vit->first));}
	      else {set_variable(vit->first,def_table->lower_bound(vit->first));}
	    }
	}
	
	/**Definition members*/
	void set_variable(Var_Id id, int v)
	{
#ifdef PRO_DEBUG
	  assert(def_table->in_range(id,v));
#endif
	  if (find(id)!=end()) {operator[](id)=v;}
	  else {Var_Val vv(id,v);insert(vv);}
	}
	
	/**Definition access members*/
	bool star_enabled(const Var_Id id) const {return def_table->star_enabled(id);}
	
	/**Inherited from Symbol.*/
	bool equal(const Symbol& S) const
	{
	  const Process* P=dynamic_cast<const Process*>(&S);
	  assert(P);
	  return equal(*P);
	}
	
	const Variable_Table* get_definition() const
	  {
	    return def_table;
	  }
	
	
	/**Inherited from Printable*/
	void print(std::ostream& out) const
	{
	  const_iterator iter=begin(),iter_end=end();
	  out << "(";
	  for(;iter!=iter_end;iter++)
	    {
	      out << iter->first << ":" << iter->second << " ";
	      //	      out << iter->second << " ";
	    }
	  out << ")";
	}
	
    public:
	
	bool equal(const Process& P) const
	{
	  const_iterator iter=begin(),iter_end=end();
	  const_iterator iter1=P.begin();
	  assert(size()==P.size());
	  
	  for(;iter!=iter_end;++iter)
	    {
	      if(def_table->star_enabled(iter->first))
		{
		  if ((iter->second != def_table->star_value(iter->first)) && (iter->second != iter1->second)) {return false;}
		  else {iter1++;}
		}
	      else
		{
		  if (iter->second!=iter1->second) {return false;}
		  else {iter1++;}
		}
	    }

	  return true;
	}
	
    private:
	const Variable_Table* def_table;
    };
  
  /**
   *class global condition.
   */
  class Global_Symbol:public Symbol,public std::map<Var_Id,int>
    {
    private:
      typedef std::map<Var_Id,int> Container;
      
    public:
      typedef Container::iterator iterator;
      typedef Container::const_iterator const_iterator;
      
    public:
      Global_Symbol() {}
      Global_Symbol(const Variable_Table* _vt):def_table(_vt) 
	{
	  for(std::map<Var_Id, Var_Def>::const_iterator vit=def_table->begin(); vit != def_table->end(); vit++)
	    {
	      /**if the star is enabled then the corresponding variable is set to the star value otherwise it is set to the lower bound**/
	      if(def_table->star_enabled(vit->first))
		{set_variable(vit->first, def_table->star_value(vit->first));}
	      else {set_variable(vit->first,def_table->lower_bound(vit->first));}
	    }
	}

	
	/**Definition members*/
	void set_variable(Var_Id id, int v)
	{
	  if(find(id)!=end())
	      erase(find(id));

	  Var_Val vv(id,v);
	  insert(vv);
	}
	
	/**Definition access members*/
	bool star_enabled(Var_Id id) {return def_table->star_enabled(id);}
	
	/**Inherited from Symbol.*/
	bool equal(const Symbol& S) const
	{ 
	  const Global_Symbol* G=dynamic_cast<const Global_Symbol*>(&S);
	  assert(G);
	  return equal(*G);
	}

	const Variable_Table* get_definition() const
	  {
	    return def_table;
	  }
	
	/**Inherited from Printable*/
	void print(std::ostream& out) const
	{
	  //	  out << true;
	  for(std::map<Var_Id, int>::const_iterator vit=begin(); vit != end(); vit++)
	    {
	      out << vit->first << ":" << vit->second << " ";
	    }
	}
	
    public:
	bool equal(const Global_Symbol& G) const
	{
	  const_iterator iter=begin(),iter_end=end();
	  const_iterator iter1=G.begin();
	  assert(size()==G.size());
	  
	  for(;iter!=iter_end;++iter)
	    {
	      if(def_table->star_enabled(iter->first))
		{
		  if ((iter->second != def_table->star_value(iter->first)) && (iter->second != iter1->second)) {return false;}
		  else {iter1++;}
		}
	      else
		{
		  if (iter->second!=iter1->second) {return false;}
		  else {iter1++;}
		}
	    }

	  return true;
	}
	
    private:
	const Variable_Table* def_table;
	
    };
  
  typedef TWord<Process> PWord;
  typedef Ref<const Global_Symbol> GSRef;
  typedef Ref<const PWord> PWRef;

}


#endif


