/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file rtransform.h
 *ADT for reversible transforms.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */
#ifndef RTRANSFORM_H
#define RTRANSFORM_H
#include <string>

#include "symbol.h"
#include "transform.h"
#include "ref.h"
#include "exconst.h"
#include "utilities.h"

//#define RTR_DEBUG
#ifdef RTR_DEBUG
#include <assert.h>
#endif

namespace rtransform
{
  using namespace transform;

  /**
   *class RTransform over symbols.
   *R stands for reversible.
   */
  template<class T>
    class RTransform:public virtual Transform<T>
  {
  public:

    /**
     * class reverse_iterator.
     * this is just an interface.
     */ 
    class reverse_iterator
    {
    public:
      typedef Ref<reverse_iterator> ref_type;
      typedef T value_type;

    public:
      virtual bool end() const=0;
      virtual reverse_iterator& operator++()=0;
      virtual reverse_iterator& operator=(const reverse_iterator& other)=0;
      virtual T operator*() const=0;
      virtual ~reverse_iterator() {}
    };

  public:
    typedef Ref<reverse_iterator> reverse_iterator_ref;
    typedef reverse_iterator iterator_type;

    reverse_iterator& begin(const Symbol& s) const
    {
      const T* r=dynamic_cast<const T*>(&s);
#ifdef RTR_DEBUG
      assert(r);
#endif
      return begin(*r); 
    }

    virtual reverse_iterator& begin(const T& s) const=0;
    virtual ~RTransform() {}
  };

  /**
   *classe reversible identity.
   */
  template<class T>
    class RIdentity:public virtual RTransform<T>,public Identity<T>
  {

  public:
    RIdentity() {}

      class reverse_iterator:public RTransform<T>::reverse_iterator
      {
      public:
	reverse_iterator() {}
	reverse_iterator(const T* _s,bool _e):ts(*_s),e(_e) {}
	  
	bool end() const {return e;}
	
	reverse_iterator& operator=(const typename RTransform<T>::reverse_iterator& other)
	  {
	    const reverse_iterator* r=dynamic_cast<const reverse_iterator*>(&other); 
	    ts=r->ts;e=r->e;
	    return *this;
	  }
	
      reverse_iterator& operator++()
	{
	  if (e) {return *this;}
	  e=true;return *this;
	}
      
      T operator*() const
      {
	return ts;
      }
      
      protected:
      T ts;
      bool e;
      }; 
      
  public:
      reverse_iterator& begin(const T& s) const
	{
	  return  *new reverse_iterator(&s,false); 
	}
      
  }; 

  /**
   *classe reversible assignement.
   *it assumes that T is of type map(string,int), that T has member functions
   *get_definition() which returns a process::variable_table
   */
  template<class T>
    class RAssignment:public virtual RTransform<T>,public Assignment<T>
  {
  private:
    typedef Assignment<T> base_type;

  public:
    RAssignment(const std::string _id, const int _val): base_type(_id,_val) {}

      class reverse_iterator:public RTransform<T>::reverse_iterator
      {
      public:
	reverse_iterator() {}
	reverse_iterator(const RAssignment<T>* _rt, const T* _s,bool _e):rt(_rt),s(*_s),e(_e)
	  {
#ifdef RTR_DEBUG
	    assert(s.get_definition());
	    assert(s.get_definition()->defined(rt->get_id()));
#endif

	    if (s[rt->get_id()]!=rt->get_val())
	      {
		if (s[rt->get_id()]!=s.get_definition()->star_value(rt->get_id())) {e=true;}
	      }
	    
	    if (s.get_definition()->star_enabled(rt->get_id()))
	      {
		i=s.get_definition()->star_value(rt->get_id());
	      }
	    else
	      {
		i=s.get_definition()->lower_bound(rt->get_id());
	      }
	  }
	  
	bool end() const {return e;}
	
	reverse_iterator& operator=(const typename RTransform<T>::reverse_iterator& other)
	  {
	    const reverse_iterator* r=dynamic_cast<const reverse_iterator*>(&other); 
	    rt=r->rt; s=r->s;e=r->e;i=r->i;
	    return *this;
	  }
	
      reverse_iterator& operator++()
	{
	  if (e) {return *this;}
	  ++i;
	  if (!(s.get_definition()->in_range(rt->get_id(),i))) {e=true;}
	  return *this;
	}
      
      T operator*() const
      {
	T temp(s);
	if (e) {return temp;}
	temp[rt->get_id()]=i;
	return temp;
      }
      
      protected:
      const RAssignment<T>* rt;
      T s;
      bool e;
      int i;
      }; 
      
  public:
      reverse_iterator& begin(const T& s) const
	{
	  return  *new reverse_iterator(this,&s,false); 
	}
      
  };
  
  template<class T,class Seq>
    class RSequence_Transform:public virtual Sequence_Transform<T,Seq>
    {
    public:
      /**
       *class reverse_iterator
       */
      class reverse_iterator
      {
      public:
	typedef ExConst<T> value_type;
	
      public:
	virtual bool end() const=0;
	virtual reverse_iterator& operator++()=0;
	virtual reverse_iterator& operator=(const RSequence_Transform<T,Seq>::reverse_iterator& other)=0;
	virtual value_type operator*() const=0;
	virtual ~reverse_iterator() {}
      };
      
    public:
      typedef Ref<reverse_iterator> reverse_iterator_ref;
      
    public: 
      virtual reverse_iterator& begin(const Seq first,const Seq last,const Seq curr) const=0;
      virtual ~RSequence_Transform() {}
    }; 
  
  template<class T,class Seq>
    class RSequence_Transform_A:public Sequence_Transform_A<T,Seq>, public virtual RSequence_Transform<T,Seq>
    {
      typedef Sequence_Transform_A<T,Seq> base_type;
      
    public:
      RSequence_Transform_A(const RTransform<T>* _t):base_type(_t) {}
	
	class reverse_iterator:public virtual RSequence_Transform<T,Seq>::reverse_iterator
	{
	public:
	  typedef typename RSequence_Transform<T,Seq>::reverse_iterator::value_type value_type;
	  
	public:
	  reverse_iterator(){}
	  reverse_iterator(const RTransform<T>* _t,const Seq _f, const Seq _l, const Seq _c, bool _e):id(new RIdentity<T>),t(_t),e(_e),f(_f),l(_l),c(_c)
	    {
	      if (!Product<RTransform<T> >().build_pro_table(table,t,&(*id),f,l,c)) {assert(!e);e=true;}
	    }
	    
	    bool end() const {return e;}
	    
	    reverse_iterator& operator++()
	    {
	      if (table.empty() || e) {e=true; return *this;}
	      if (!Product<RTransform<T> >().update(table,table.size()-1)) {e=true;}
	      return *this;
	    }
	    
	    reverse_iterator& operator=(const typename RSequence_Transform<T,Seq>::reverse_iterator& other)
	      {
		const reverse_iterator* r=dynamic_cast<const reverse_iterator*>(&other);
		assert(r);
		e=r->e;table=r->table;l=r->l;c=r->c;f=r->f;t=r->t;id=r->id;
		return *this;
	      }
	    
	    value_type operator*() const
	    {
	      TWord<T> w;
	      unsigned int i=0;
	      for( ; i!=table.size(); ++i){w.push_back(**(table[i].first.second));}
	      value_type res(w);
	      res.adjust(f,c,l);
	      return res;
	    }

	protected:
	    Ref<RIdentity<T> > id;
	    const RTransform<T>* t;
	    typename Product<RTransform<T> >::vec_type table;
	    bool e;
	    Seq f;
	    Seq l;
	    Seq c;
	};
      
    public:
      reverse_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  const RTransform<T>* r=dynamic_cast<const RTransform<T>*>(base_type::get_trans());
	  assert(r);
	  return *new reverse_iterator(r,first,last,curr,false);
	}
    };
}

#endif
