/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file starpredicate.h
 *ADT for predicates symbols and constraints.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef STARPREDICATE_H
#define STARPREDICATE_H
#include <vector>

#include "predicate.h"
#include "symbol.h"
#include "ref.h"
#include "my_functional.h"

//#define MPR_DEBUG
#ifdef MPR_DEBUG
#include <assert>
#endif

namespace starpredicate
{  
  using namespace predicate;
  using namespace process;
  using namespace my_stl;

  /**
   *class mpredicates.
   *it is parameterized by T which represents a concret instantiation of symbol
   */
  template<class T>
    class StarPredicate:public virtual Predicate
  {
  public:
    /**
     *class constraint_iterator to iterate the set of symbols which may or may not make the 
     *predicate hold.
     */
    class constraint_iterator
    {
    public:
      typedef Ref<constraint_iterator> ref_type;
      typedef T value_type;

    public:
      virtual bool end() const=0;
      virtual constraint_iterator& operator++()=0;
      virtual constraint_iterator& operator=(const StarPredicate<T>::constraint_iterator& other)=0;
      virtual T operator*() const=0;
      virtual ~constraint_iterator() {}
    };
    
  public:
    typedef Ref<constraint_iterator> constraint_iterator_ref;
    typedef constraint_iterator iterator_type;

  public:

    constraint_iterator& begin(const Symbol& s) const
      {
	const T* r=dynamic_cast<const T*>(&s);
#ifdef SPR_DEBUG
	assert(r);
#endif
	return begin(*r); 
      }

    virtual StarPredicate<T>& negate() const=0; 

    virtual constraint_iterator& begin(const T& s) const=0;

    virtual ~StarPredicate() {}
  };

/**
   *class makehold functions.
   *it is parameterized by T which represents a concret instantiation of symbol
   */
  template<class T>
    class StarFunction:public virtual Function
  {
  public:
    /**
     *class constraint_iterator
     */
    class constraint_iterator
    {
    public:
      virtual bool end() const=0;
      virtual constraint_iterator& operator++()=0;
      virtual constraint_iterator& operator=(const StarFunction<T>::constraint_iterator& other)=0;
      virtual T operator*() const=0;
      virtual ~constraint_iterator() {}
    };
    
  public:
    typedef Ref<constraint_iterator> constraint_iterator_ref;
    
  public:

    constraint_iterator& begin(const Symbol& s) const
      {
	const T* r=dynamic_cast<const T*>(&s);
#ifdef SPR_DEBUG
	assert(r);
#endif
	return begin(*r); 
      }

    virtual constraint_iterator& begin(const T& s) const=0;

    virtual StarFunction<T>& negate() const=0;

    virtual ~StarFunction() {}
  };

  /**
   *class makehold composed predicates.
   *it is parameterized by T which represents a concret instantiation of symbol
   *and the type of the sub_predicates.
   */
  template<class Pred,class T>
    class StarComposed_Predicate:public virtual StarPredicate<T>,public virtual Composed_Predicate<Pred>
  {
  public:
    virtual ~StarComposed_Predicate() {}
  };

  /**
   *the make hold true over T should return all possible instantiation of
   *type T. Althgough the predicate true is universal (doesn't need to be 
   *parameterized) the mpredicate version needs a domain which is 
   *provided by T.
   */
  template<class T>
    class StarTrue:public virtual StarPredicate<T>, public True
  {
  public:
    StarTrue() {}

    class constraint_iterator:public StarPredicate<T>::constraint_iterator
    {
    public:
      constraint_iterator() {}
      /**needs a valid type T to get the variable definitions*/
      constraint_iterator(const T* s,bool _e):ts(*s),e(_e){}

	bool end() const {return e;}
	
	constraint_iterator& operator++()
	  {
	    if (e){return *this;}
	    e=true;return *this;
	  }
	
	constraint_iterator& operator=(const typename StarPredicate<T>::constraint_iterator& other)
	  {
	    const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	    ts=r->ts;e=r->e;
	    return *this;
	  }
	
	T operator*() const
	{
	  return ts;
	}    

    protected:
      T ts;
      bool e;
    };

  public:

    constraint_iterator& begin(const T& s) const
      {
	return *new constraint_iterator(&s,false);
      } 

    StarPredicate<T>& negate() const;

  };
template<class T>
    class StarFalse:public virtual StarPredicate<T>, public True
  {
  public:
    StarFalse() {}

    class constraint_iterator:public StarPredicate<T>::constraint_iterator
    {
    public:
      constraint_iterator() {}
      /**needs a valid type T to get the variable definitions*/
      constraint_iterator(const T* s,bool _e):ts(s->get_definition()),e(_e) {}

	bool end() const {return true;}
	
	constraint_iterator& operator++()
	  {
	    return *this;
	  }
	
	constraint_iterator& operator=(const typename StarPredicate<T>::constraint_iterator& other)
	  {
	    return *this;
	  }
	
	T operator*() const
	{
	  return ts;
	}

    protected:
      T ts;
      bool e;
    };

  public:

    constraint_iterator& begin(const T& s) const
      {
	return *new constraint_iterator(&s,true);
      } 

    StarPredicate<T>& negate() const;

  };

  /**
   *make hold strategy for intersection.
   */
  template<class Pred,class T>
    class Intersection_Strategy
  {
    typedef StarComposed_Predicate<Pred,T> pred_type;
    typedef typename Pred::constraint_iterator_ref sub_type; 

  public:
    Intersection_Strategy() {}

    Intersection_Strategy(const pred_type* _pr,const T* _s,bool _e):pr(_pr),s(*_s),e(_e)
      {
	//if (!(pr->holds_for_subset(s))) {e=true;}
	if (!(pr->operator()(s))) {e=true;}

	s1=sub_type(&pr->get_left()->begin(s));
	s2=sub_type(&pr->get_right()->begin(**s1));
	
	//if ((*s1).end() && (*s2).end()) {e=true;}
	
	if (!e) {s=**s2;}

	while (!e && !pr->operator()(s)){operator++();}	
	
      }

    bool end() const { return e;}

    bool operator++()
      {
	if (e) {return e;}
	else
	  {
	    ++(*s2);
	    if ((*s2).end())
	      {
		++(*s1);
		if ((*s1).end()) {e=true;return e;}
		else {s2=sub_type(&pr->get_right()->begin(**s1));s=**s2;}
	      }
	    else {s=**s2;}
	    while (!e && !pr->operator()(s)){return operator++();}
	    return e;
	  }

      }  

    bool operator=(const typename StarPredicate<T>::constraint_iterator& other)
      {
	const Intersection_Strategy<Pred,T>* r=dynamic_cast<const Intersection_Strategy<Pred,T>*>(&other);
	pr=r->pr;s1=r->s1;s2=r->s2;e=r->e;s=r->s;
	return true;
      }

    T operator*() const {return s;}

  private:
    const pred_type* pr;
    sub_type s1;
    sub_type s2;
    T s;
    bool e;
  };

 /**
   *make hold strategy for union.
   */
  template<class Pred,class T=Symbol>
    class Union_Strategy
  {
    typedef StarComposed_Predicate<Pred,T> pred_type;
    typedef typename Pred::constraint_iterator_ref sub_type;

  public:
    Union_Strategy() {}

    Union_Strategy(const pred_type* _pr,const T* _s,bool _e):pr(_pr),s(*_s),e(_e)
      {
	//if (!(pr->holds_for_subset(s))) {e=true;}
	if (!(pr->operator()(s))) {e=true;}
	      
	s1=sub_type(&pr->get_left()->begin(s));
	s2=sub_type(&pr->get_right()->begin(s));
	
	if (!e) {s=**s1;}
	

	//if ((*s1).end() && (*s2).end()) {e=true;}

	if (!e && (*s1).end()) {s=**s2;}
	
	while (!e && !pr->operator()(s)){operator++();}
      }

    bool end() const { return e;}

    bool operator++()
      {
	if (e) {return e;}
	if ((*s1).end())
	  {
	    ++(*s2);
	    if ((*s2).end()) {e=true; return e;}
	    else {s=**s2;}
	  }
	else
	  {
	    ++(*s1);
	    if ((*s1).end()) 
	      {
		if ((*s2).end()) {e=true; return e;}
		else {s=**s2;}
	      }
	    else {s=**s1;}
	  }
	while (!e && !pr->operator()(s)){return operator++();}
	return e;
      }  

    bool operator=(const typename StarPredicate<T>::constraint_iterator& other)
      {
	const Union_Strategy<Pred,T>* r=dynamic_cast<const Union_Strategy<Pred,T>*>(&other);
	pr=r->pr;s1=r->s1;s2=r->s2;e=r->e;s=r->s;
	return true;
      }

    T operator*() const {return s;}

  private:
    const pred_type* pr;
    sub_type s1;
    sub_type s2;
    T s;
    bool e;
  };

  /**
   *mpredicates of the form "variable op constant" where op in {==,!=,<,>,<=,>=}.
   *the constraint_iterator ranges over all possible instantion of the variable until the 
   *predicate hols. !!! it is the naive approach
   */
  template<class T>
    class StarIntOpPredicate:public virtual StarComposed_Predicate<StarFunction<T>,T>, private OpPredicate<StarFunction<T> >
  { 
    private:
    typedef OpPredicate<StarFunction<T> > base_type;
    
    public:
    StarIntOpPredicate(const StarFunction<T>* _p1, const StarFunction<T>* _p2, const typename base_type::op_type* _op):base_type(_p1,_p2,_op), p1(_p1), p2(_p2) {}

    bool holds_for_subset(const Symbol& s) const 
    { 
      return get_left()->holds_for_subset(s) || get_right()->holds_for_subset(s); 
    }
    
    bool operator()(const Symbol& s) const 
    { 
      return base_type::get_op()->operator()(get_left()->operator()(s),get_right()->operator()(s)) || holds_for_subset(s); 
    }
    
    StarPredicate<T>& negate() const
      {
	return *new StarIntOpPredicate(get_left(),get_right(),&base_type::get_op()->negate());
      }

    const StarFunction<T>* get_left() const { return base_type::get_left(); }
    
    const StarFunction<T>* get_right() const { return base_type::get_right(); }

    void print(std::ostream& out) const {base_type::print(out);}
    
    class constraint_iterator:public StarPredicate<T>::constraint_iterator,private Intersection_Strategy<StarFunction<T>,T>
    {
      typedef StarComposed_Predicate<StarFunction<T>,T> pred_type;
      typedef Intersection_Strategy<StarFunction<T>,T> base_type;

      public:
      constraint_iterator() {}

      constraint_iterator(const pred_type* _pr,const T* _s,bool _e):base_type(_pr,_s,_e) {}
      
      bool end() const { return base_type::end(); }
      
      constraint_iterator& operator++()
      {
	base_type::operator++();
	return *this;
      }
      
      constraint_iterator& operator=(const typename StarPredicate<T>::constraint_iterator& other)
      {
	base_type::operator=(other);
	return *this;
      }
      
      T operator*() const { return base_type::operator*(); }

    };
      
    constraint_iterator& begin(const T& s) const
    {
      return *new constraint_iterator(this,&s,false);
    }	

  private:
    const StarFunction<T>* p1;
    const StarFunction<T>* p2;
  };

  /**
   *mpredicates of the form "mpredicate && mpredicate".
   */
  template<class T>
    class StarAndPredicate:public StarComposed_Predicate<StarPredicate<T>,T> , private OpPredicate<StarPredicate<T> >
    {
    private:
      typedef OpPredicate<StarPredicate<T> > base_type;
      
    public:
      StarAndPredicate(const StarPredicate<T>* _p1, const StarPredicate<T>* _p2):base_type(_p1,_p2,new And()){}
	
	bool operator()(const Symbol& s) const 
	{ 
	  return base_type::get_op()->operator()(get_left()->operator()(s),get_right()->operator()(s));
	}
	
	bool holds_for_subset(const Symbol& s) const 
	{ 
	  return (get_left()->holds_for_subset(s) || get_right()->holds_for_subset(s)) && operator()(s);
	}

	StarPredicate<T>& negate() const;
	
	const StarPredicate<T>* get_left() const { return base_type::get_left(); }
	
	const StarPredicate<T>* get_right() const { return base_type::get_right(); }
	
	void print(std::ostream& out) const {base_type::print(out);}

	class constraint_iterator:public StarPredicate<T>::constraint_iterator,private Intersection_Strategy<StarPredicate<T>,T>
	  {
	    typedef StarComposed_Predicate<StarPredicate<T>,T> pred_type;
	    typedef Intersection_Strategy<StarPredicate<T>,T> base_type;
	    
	  public:
	    constraint_iterator() {}
	    
	    constraint_iterator(const pred_type* _pr,const T* _s,bool _e):base_type(_pr,_s,_e) {}
	      
	      bool end() const { return base_type::end(); }
	      
	      constraint_iterator& operator++()
		{
		  base_type::operator++();
		  return *this;
		}
	      
	      constraint_iterator& operator=(const typename StarPredicate<T>::constraint_iterator& other)
		{
		  base_type::operator=(other);
		  return *this;
		}
	      
	      T operator*() const { return base_type::operator*(); }
	  };
	
	constraint_iterator& begin(const T& s) const
	  {
	    return *new constraint_iterator(this,&s,false);
	  }
    };

  /**
   *mpredicates of the form "mpredicate || mpredicate".
   */
  template<class T>
    class StarOrPredicate:public StarComposed_Predicate<StarPredicate<T>,T>, private OpPredicate<StarPredicate<T> >
    {
    private:
      typedef OpPredicate<StarPredicate<T> > base_type;
      
    public:
      StarOrPredicate(const StarPredicate<T>* _p1, const StarPredicate<T>* _p2):base_type(_p1,_p2,new Or()){}
	
	bool operator()(const Symbol& s) const 
	{ 
	  return base_type::get_op()->operator()(get_left()->operator()(s),get_right()->operator()(s));
	}

	bool holds_for_subset(const Symbol& s) const 
	{ 
	  return (get_left()->holds_for_subset(s) && !get_right()->operator()(s)) || (get_right()->holds_for_subset(s) && !get_left()->operator()(s)) || (get_right()->holds_for_subset(s) && get_left()->holds_for_subset(s));
	}

	StarPredicate<T>& negate() const;

	const StarPredicate<T>* get_left() const { return base_type::get_left(); }
    
	const StarPredicate<T>* get_right() const { return base_type::get_right(); }

	 void print(std::ostream& out) const {base_type::print(out);}

	class constraint_iterator:public StarPredicate<T>::constraint_iterator,private Union_Strategy<StarPredicate<T>,T>
	  {
	    typedef StarComposed_Predicate<StarPredicate<T>,T> pred_type;
	    typedef Union_Strategy<StarPredicate<T>,T> base_type;
	    
	  public:
	    constraint_iterator() {}
	    
	    constraint_iterator(const pred_type* _pr,const T* _s,bool _e):base_type(_pr,_s,_e) {}
	      
	      bool end() const { return base_type::end(); }
	      
	      constraint_iterator& operator++()
		{
		  base_type::operator++();
		  return *this;
		}
	      
	      constraint_iterator& operator=(const typename StarPredicate<T>::constraint_iterator& other)
		{
		  base_type::operator=(other);
		  return *this;
		}
	      
	      T operator*() const { return base_type::operator*(); }
	      
	  };

	constraint_iterator& begin(const T& s) const
	  {
	    return *new constraint_iterator(this,&s,false);
	  }

    };
  
  template<class T>
    StarPredicate<T>& StarAndPredicate<T>::negate() const
    {
      return *new StarOrPredicate<T>(&get_left()->negate(),&get_right()->negate());
    }

  template<class T>
    StarPredicate<T>& StarOrPredicate<T>::negate() const
    {
      return *new StarAndPredicate<T>(&get_left()->negate(),&get_right()->negate());
    }

  template<class T>
    StarPredicate<T>& StarTrue<T>::negate() const
    {
      return *new StarFalse<T>();
    }

  template<class T>
    StarPredicate<T>& StarFalse<T>::negate() const
    {
      return *new StarTrue<T>();
    }

}

#endif
