/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file mspredicate.h
 *ADT for sequence (exists forall) predicates which can be made "hold".
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef STARQUANTIFIER_H
#define STARQUANTIFIER_H

#include "predicate.h"
#include "starpredicate.h"
#include "creator.h"
#include "exconst.h"
#include "utilities.h"

#include "tword.h"
#include "tword_op.h"
#include "ref.h"

namespace starquantifier
{
  using namespace creator;
  using namespace predicate;
  using namespace starpredicate;

  template<class T,class Seq>
    class StarQuantifier:public virtual Quantifier<Seq>
  {
  public:
    /**
     *class constraint_iterator
     */
    class constraint_iterator
    {
    public:
      typedef ExConst<T> value_type;
      typedef Ref<constraint_iterator> ref_type;

    public:
      virtual bool end() const=0;
      virtual constraint_iterator& operator++()=0;
      virtual constraint_iterator& operator=(const StarQuantifier<T,Seq>::constraint_iterator& other)=0;
      virtual value_type operator*() const=0;
      virtual ~constraint_iterator() {}
    };

  public:
    typedef Ref<constraint_iterator> constraint_iterator_ref;
    typedef constraint_iterator iterator_type;

  public: 
    virtual StarQuantifier<T,Seq>& negate() const=0;
    virtual constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const=0;
    virtual ~StarQuantifier() {}
    
  };

  template<class T,class Seq>
    class StarForAllStrategy
  {
    typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;

  public:
    StarForAllStrategy() {}
    StarForAllStrategy(const StarPredicate<T>* _p,const Seq _f, const Seq _l, const Seq _c, bool _e):id(new StarTrue<T>()),p(_p),e(_e),f(_f),l(_l),c(_c) 
      {
	if (!Product<StarPredicate<T> >().build_pro_table(table,p,&(*id),f,l,c)) {e=true;}

      }
      //      ~StarForAllStrategy() {delete(p);}
      
      bool end() const { return e;}
      
      bool operator++()
	{
	  if (table.empty() || e) {e=true; return e;}
	  if (!Product<StarPredicate<T> >().update(table,table.size()-1)) {e=true;}
	  return e;
	}

      bool operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	{
	  const StarForAllStrategy<T,Seq>* r=dynamic_cast<const StarForAllStrategy<T,Seq>*>(&other);
	  assert(r);
	  return operator=(*r);
	}

      bool operator=(const StarForAllStrategy<T,Seq>& other)
	{
	  id=other.id;p=other.p;e=other.e;table=other.table;f=other.f;l=other.l;c=other.c;
	  return true;
	}
      
      
      value_type operator*() const
	{
	  TWord<T> w;
	  unsigned int i=0;
	  for( ; i!=table.size(); ++i){w.push_back(**(table[i].first.second));}
	  value_type res(w);
	  return res;
	}
     
  protected:
      Ref<StarTrue<T> > id;
      const StarPredicate<T>* p;
      typename Product<StarPredicate<T> >::vec_type table;
      bool e;
      Seq f;
      Seq l;
      Seq c;
  };


  template<class T,class Seq>
    class StarForAllA:public virtual StarQuantifier<T,Seq>,public ForAllA<Seq,StarPredicate<T> >
  {
  private:
    typedef ForAllA<Seq,StarPredicate<T> > base_type;

  public:
    StarForAllA(const StarPredicate<T>* _p):base_type(_p) {}

      class constraint_iterator:public StarQuantifier<T,Seq>::constraint_iterator,private StarForAllStrategy<T,Seq>
	{
	  typedef StarForAllStrategy<T,Seq> base_type;
	  typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;

	public:
	  constraint_iterator() {}
	  constraint_iterator(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type(_p,_f,_l,_c,_e),f(_f),c(_c),l(_l){}

	    bool end() const { return base_type::end();}

	    constraint_iterator& operator++()
	      {
		base_type::operator++();
		return *this;
	      }

	    constraint_iterator& operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	    {
	      base_type::operator=(other);
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      assert(r);
	      c=r->c;f=r->f;l=r->l;
	      return *this;
	    }

	
	    value_type operator*() const
	      {
		value_type res(base_type::operator*());
		res.adjust(f,c,l);
		return res;
	      }

	protected:
	    Seq f;
	    Seq c;
	    Seq l;
	};

    StarQuantifier<T,Seq>& negate() const;

    constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  return *new constraint_iterator(base_type::get_pred(),first,last,curr,!holds_for_subset(first,last,curr));
	}
  };
  
  template<class T,class Seq>
    class StarForAllR:public virtual StarQuantifier<T,Seq>,public ForAllR<Seq,StarPredicate<T> >
  {
  private:
    typedef ForAllR<Seq,StarPredicate<T> > base_type;

  public:
    StarForAllR(const StarPredicate<T>* _p):base_type(_p) {}

      class constraint_iterator:public StarQuantifier<T,Seq>::constraint_iterator,private StarForAllStrategy<T,Seq>
	{
	  typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
	  typedef StarForAllStrategy<T,Seq> base_type;

	public:
	  constraint_iterator() {}
	  constraint_iterator(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type(_p,_c,_l,_c,_e),f(_f),c(_c),l(_l) {/*Seq t=_c;++t; base_type=StarForAllStrategy<T,Seq>(_p,t,_l,_e);*/}

	    bool end() const { return base_type::end();}

	    constraint_iterator& operator++()
	      {
		base_type::operator++();
		return *this;
	      }

	    constraint_iterator& operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	    {
	      base_type::operator=(other);
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      assert(r);
	      c=r->c;f=r->f;l=r->l;/*base_type=r->base_type;*/
	      return *this;
	    }

	    value_type operator*() const
	      {
		TWord<T> w;
		w.insert(w.begin(),f,c);
		//w.push_back(*c);
		TWord<T> w1(base_type::operator*().word);
		w.merge(w1);
		value_type res(w);
		res.adjust(f,c,l);
		return res;
	      }


	protected:
	    /*StarForAllStrategy<T,Seq> base_type;*/
	    Seq f;
	    Seq c;
	    Seq l;    
	};
  
    StarQuantifier<T,Seq>& negate() const;

      constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  return *new constraint_iterator(base_type::get_pred(),first,last,curr,!holds_for_subset(first,last,curr));
	}
  };
  
  template<class T,class Seq>
    class StarForAllL:public virtual StarQuantifier<T,Seq>,public ForAllL<Seq,StarPredicate<T> >
  {
  private:
    typedef ForAllL<Seq,StarPredicate<T> > base_type;

  public:
    StarForAllL(const StarPredicate<T>* _p):base_type(_p) {}

      class constraint_iterator:public StarQuantifier<T,Seq>::constraint_iterator,private StarForAllStrategy<T,Seq>
	{
	  typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
	  typedef StarForAllStrategy<T,Seq> base_type;

	public:
	  constraint_iterator() {}
	  constraint_iterator(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type(_p,_f,_c,_c,_e),f(_f),c(_c),l(_l){}

	    bool end() const { return base_type::end();}

	    constraint_iterator& operator++()
	      {
		base_type::operator++();
		return *this;
	      }

	    constraint_iterator& operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	    {
	      base_type::operator=(other);
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      assert(r);
	      c=r->c;f=r->f;l=r->l;
	      return *this;
	    }

	
	    value_type operator*() const
	      {
		TWord<T> w;
		w.insert(w.begin(),c,l);
		TWord<T> w1(base_type::operator*().word);
		w1.merge(w);
		value_type res(w1);
		res.adjust(f,c,l);
		return res;
	      }

	protected:
	    Seq f;
	    Seq c;
	    Seq l;    
	};

    StarQuantifier<T,Seq>& negate() const;

      constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  return *new constraint_iterator(base_type::get_pred(),first,last,curr,!holds_for_subset(first,last,curr));
	}
  };

  template<class T,class Seq>
    class StarItExistsStrategy;

  template<class T,class Seq>
    class StarExistsPhase_1
  {
  private:
    typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
    /*typedef typename StarPredicate<T>::constraint_iterator_ref subtype;
      typedef std::pair<subtype,T> StarH_R;*/

  public:
    StarExistsPhase_1() {}
    StarExistsPhase_1(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):id(new StarFalse<T>()),p(_p),e(_e),f(_f),l(_l),c(_c) 
      {/*
	Seq t=f;
	bool check=false;
	int i=0;
	while(t!=l)
	  {
	    table.push_back(StarH_R(subtype(&(p->begin(*t))),*t)); 
	    if (!(*(table[i].first)).end()) {check=true;}
	    ++i;
	    ++t;
	  }*/
	if (!Sum<StarPredicate<T> >().build_sum_table(table,p,&(*id),f,l,c)) {e=true;}
      }
      
      bool end() const { return e;}

      bool operator++()
      {
	if (table.empty()) {e=true; return e;}
	if (!Sum<StarPredicate<T> >().update(table,table.size()-1,false)) {e=true;}
	return e;
      }

      bool operator=(const StarItExistsStrategy<T,Seq>& other);
	
      value_type operator*() const
	{
	  TWord<T> w;
	  unsigned int i=table.size();
	  bool push=false;
	  
	  for( ; i!=0; --i)
	    {if (!push && !e)
		{if (!(*table[i-1].first).end())
		    {w.push_front(**(table[i-1].first));push=true;}
		  else {w.push_front(table[i-1].second);}
		}
	      else {w.push_front(table[i-1].second);}}

	  value_type res(w);
	  return res;
	}
      
  protected:
      //first phase: when the quantifier holds du to few star constraints in the given configuration.
      Ref<StarFalse<T> > id;
      const StarPredicate<T>* p;
      typename Sum<StarPredicate<T> >::vec_type table;
      bool e;
      Seq f;
      Seq l;
      Seq c;
  };

  template<class T,class Seq>
    class StarExistsPhase_2
  {
  private:
    typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
    typedef Ref<StarPredicate<T> > pred_ptr;

  public:
    StarExistsPhase_2() {}
    StarExistsPhase_2(const StarPredicate<T>* _p, const Seq _f, const Seq _l, const Seq _c, bool _e):ptr(pred_ptr(&_p->negate())),base_type(StarForAllStrategy<T,Seq>(&(*ptr),_f,_l,_c,_e)),e(_e),flag(_f==_c)
      { 
	curr_w=Ref<TWord<T> >(new TWord<T>(base_type.operator*().word));
	p=Creator<T>(_p); 
	p_iter=p.begin();
	w_iter=gen.start(*p_iter,curr_w->begin(),curr_w->end());
	if (!e && flag) {++w_iter;}
      }
     
      bool end() const {return e;}

      bool operator++()
      {
	if (e==true) {return e;}
	if (!update()) {e=true;}
	return e;
      }

      bool operator=(const StarItExistsStrategy<T,Seq>& other);

      value_type operator*() const
	{
	  if (e) {return value_type(*curr_w);}
	  else {return value_type(*w_iter);}
	}
      

  private:
      bool update()
      {
	if (e) {return false;}
	//if (p_iter.end()) {e=true;return false;}
	else
	{
	  ++w_iter;
	  if (w_iter.end()) 
	    {
	      ++p_iter;//std::cout << "allo " << *p_iter << std::endl;
	      if (p_iter.end())
		{
		  base_type.operator++();
		  if (base_type.end()) {e=true;return false;}
		  else 
		    {
		      curr_w=Ref<TWord<T> >(new TWord<T>(base_type.operator*().word));
		      p_iter=p.begin();
		      w_iter=gen.start(*p_iter,curr_w->begin(),curr_w->end());
		      if (!e && flag) {++w_iter;}
		      return true;
		    }
		}
	      else
		{
		  w_iter=gen.start(*p_iter,curr_w->begin(),curr_w->end());
		  if (!e && flag) {++w_iter;}
		  return true;
		}
	    }
	  else {return true;}
	}
      }

  protected:
      pred_ptr ptr;
      StarForAllStrategy<T,Seq> base_type;
      Creator<T> p;
      typename Creator<T>::iterator p_iter;
      Ref<TWord<T> > curr_w;
      TGenerator<T> gen;
      typename TGenerator<T>::IWR w_iter;
      bool e;
      bool flag;
  };

  template<class T,class Seq>
    class StarItExistsStrategy:public StarExistsPhase_1<T,Seq>,public StarExistsPhase_2<T,Seq>
  {
  private:
    typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
    typedef StarExistsPhase_1<T,Seq> base_type_1;
    typedef StarExistsPhase_2<T,Seq> base_type_2;

  public:
    StarItExistsStrategy() {}
    StarItExistsStrategy(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type_1(_p,_f,_l,_c,_e),base_type_2(_p,_f,_l,_c,_e){}
      
      bool end() const { return base_type_1::end() && base_type_2::end();}
      
      bool operator++()
	{
	  if (base_type_1::end() && base_type_2::end()) {return false;}
	  if (!base_type_1::end())
	    { 
	      base_type_1::operator++();
	      if (base_type_1::end()) {return base_type_2::end();}
	      else {return true;}
	    }
	  else {base_type_2::operator++();}
	  return base_type_2::end();
	}
	
      bool operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	{
	  const StarItExistsStrategy<T,Seq>* r=dynamic_cast<const StarItExistsStrategy<T,Seq>*>(&other);
	  assert(r);
	  return operator=(*r);
	}

      bool operator=(const StarItExistsStrategy<T,Seq>& other)
	{
	  base_type_1::operator=(other);
	  base_type_2::operator=(other);
	  return true;
	}
      
      value_type operator*() const
	{
	  if (base_type_1::end() && base_type_2::end()) {return base_type_1::operator*();}
	  if (!base_type_1::end() || base_type_2::end()) {return base_type_1::operator*();}
	  else {return base_type_2::operator*();}
	}
  };

template<class T,class Seq>
    class StarItExistsA:public virtual StarQuantifier<T,Seq>,public ItExistsA<Seq,StarPredicate<T> >
  {
  private:
    typedef ItExistsA<Seq,StarPredicate<T> > base_type;

  public:
    StarItExistsA(const StarPredicate<T>* _p):base_type(_p) {}

      class constraint_iterator:public StarQuantifier<T,Seq>::constraint_iterator,private StarItExistsStrategy<T,Seq>
	{
	  typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
	  typedef StarItExistsStrategy<T,Seq> base_type;

	public:
	  constraint_iterator() {}
	  constraint_iterator(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type(_p,_f,_l,_c,_e),f(_f),c(_c),l(_l) {}

	    bool end() const { return base_type::end();}

	    constraint_iterator& operator++()
	      {
		base_type::operator++();
		return *this;
	      }

	    constraint_iterator& operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	    {
	      base_type::operator=(other);
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      assert(r);
	      l=r->l;f=r->f;c=r->c;
	      return *this;
	    }
	
	    value_type operator*() const
	      {
		value_type res(base_type::operator*());
		res.adjust(f,c,l);
		return res;
	      }

	protected:
	    Seq f;
	    Seq c;
	    Seq l;
	};

    StarQuantifier<T,Seq>& negate() const;

      constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  return *new constraint_iterator(base_type::get_pred(),first,last,curr,!holds_for_subset(first,last,curr));
	}
  };

  template<class T,class Seq>
    class StarItExistsR:public virtual StarQuantifier<T,Seq>,public ItExistsR<Seq,StarPredicate<T> >
  {
  private:
    typedef ItExistsR<Seq,StarPredicate<T> > base_type;

  public:
    StarItExistsR(const StarPredicate<T>* _p):base_type(_p) {}

      class constraint_iterator:public StarQuantifier<T,Seq>::constraint_iterator,private StarItExistsStrategy<T,Seq>
	{
	private:
	  typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
	  typedef StarItExistsStrategy<T,Seq> base_type;

	public:
	  constraint_iterator() {}
	  constraint_iterator(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type(_p,_c,_l,_c,_e),f(_f),c(_c),l(_l) 
	    {/* Seq t=_c;++t;base_type=StarItExistsStrategy<T,Seq>(_p,t,_l,_e); */}

	    bool end() const { return base_type::end();}

	    constraint_iterator& operator++()
	      {
		base_type::operator++();
		return *this;
	      }

	    constraint_iterator& operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	    {
	      base_type::operator=(other);
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      assert(r);
	      c=r->c;f=r->f;l=r->l;
	      return *this;
	    }

	    value_type operator*() const
	      {
		TWord<T> w;
		w.insert(w.end(),f,c);
		//w.push_back(*c);
		TWord<T> w1(base_type::operator*().word);
		w.merge(w1);
		value_type res(w);
		res.adjust(f,c,l);
		return res;
	      }

	protected:
	    /*StarItExistsStrategy<T,Seq> base_type;*/
	    Seq f;
	    Seq c;
	    Seq l;    
	};

    StarQuantifier<T,Seq>& negate() const;

      constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  return *new constraint_iterator(base_type::get_pred(),first,last,curr,!holds_for_subset(first,last,curr));
	}
  };

  template<class T,class Seq>
    class StarItExistsL:public virtual StarQuantifier<T,Seq>,public ItExistsL<Seq,StarPredicate<T> >
  {
  private:
    typedef ItExistsL<Seq,StarPredicate<T> > base_type;

  public:
    StarItExistsL(const StarPredicate<T>* _p):base_type(_p) {}

      class constraint_iterator:public StarQuantifier<T,Seq>::constraint_iterator,private StarItExistsStrategy<T,Seq>
	{
	  typedef typename StarQuantifier<T,Seq>::constraint_iterator::value_type value_type;
	  typedef StarItExistsStrategy<T,Seq> base_type;

	public:
	  constraint_iterator() {}
	  constraint_iterator(const StarPredicate<T>* _p,const Seq _f, const Seq _l,const Seq _c,bool _e):base_type(_p,_f,_c,_c,_e),f(_f),c(_c),l(_l)
	    {
	      count=0;
	      Seq it=f;
	      while(it!=c){++it;++count;}
	    }

	    bool end() const { return base_type::end();}

	    constraint_iterator& operator++()
	      {
		base_type::operator++();
		return *this;
	      }

	    constraint_iterator& operator=(const typename StarQuantifier<T,Seq>::constraint_iterator& other)
	    {
	      base_type::operator=(other);
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      assert(r);
	      c=r->c;f=r->f;l=r->l;count=r->count;
	      return *this;
	    }

	
	    value_type operator*() const
	      {
		TWord<T> w;
		w.insert(w.begin(),c,l);
		TWord<T> w1(base_type::operator*().word);
		w1.merge(w);
		value_type res(w1);
		res.adjust(f,c,l);
		if (count > w1.size() && c!=l) {++res.curr;}
		return res;
	      }

	protected:
	    unsigned int count;
	    Seq f;
	    Seq c;
	    Seq l;    
	};

    StarQuantifier<T,Seq>& negate() const;

      constraint_iterator& begin(const Seq first,const Seq last,const Seq curr) const
	{
	  return *new constraint_iterator(base_type::get_pred(),first,last,curr,!holds_for_subset(first,last,curr));
	}
  };

  template<class T,class Seq>
    bool StarExistsPhase_1<T,Seq>::operator=(const StarItExistsStrategy<T,Seq>& other)
    {
      const StarExistsPhase_1<T,Seq>* r=dynamic_cast<const StarExistsPhase_1<T,Seq>* >(&other);
      assert(r);
      id=r->id;p=r->p;e=r->e;table=r->table;f=r->f;l=r->l;c=r->c;
      return true;
    }

  template<class T,class Seq>
    bool StarExistsPhase_2<T,Seq>::operator=(const StarItExistsStrategy<T,Seq>& other)
    {
      const StarExistsPhase_2<T,Seq>* r=dynamic_cast<const StarExistsPhase_2<T,Seq>* >(&other);
      assert(r);
      p=r->p;p_iter=r->p_iter;gen=r->gen;curr_w=r->curr_w;w_iter=r->w_iter;e=r->e;base_type=r->base_type;ptr=r->ptr;flag=r->flag;
      return true;
    }

  template<class T,class Seq>
    StarQuantifier<T,Seq>& StarForAllA<T,Seq>::negate() const
    {
      return *new StarItExistsA<T,Seq>(&base_type::get_pred()->negate());
    }

  template<class T,class Seq>
    StarQuantifier<T,Seq>& StarForAllR<T,Seq>::negate() const
    {
      return *new StarItExistsR<T,Seq>(&base_type::get_pred()->negate());
    }

  template<class T,class Seq>
    StarQuantifier<T,Seq>& StarForAllL<T,Seq>::negate() const
    {
      return *new StarItExistsL<T,Seq>(&base_type::get_pred()->negate());
    }

  template<class T,class Seq>
    StarQuantifier<T,Seq>& StarItExistsA<T,Seq>::negate() const
    {
      return *new StarForAllA<T,Seq>(&base_type::get_pred()->negate());
    }

  template<class T,class Seq>
    StarQuantifier<T,Seq>& StarItExistsR<T,Seq>::negate() const
    {
      return *new StarForAllR<T,Seq>(&base_type::get_pred()->negate());
    }

  template<class T,class Seq>
    StarQuantifier<T,Seq>& StarItExistsL<T,Seq>::negate() const
    {
      return *new StarForAllL<T,Seq>(&base_type::get_pred()->negate());
    }

}

#endif





