/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file task.h
 *ADT for taskss.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef TASK_H
#define TASK_H

#include "process.h"
#include "transform.h"
#include "rtransform.h"
#include "ref.h"
#include "printable.h"
#include "exconst.h"

#include <assert.h>

//#define TAS_DEBUG

namespace action
{
  /**
   *Typedef for the parameter type of condition. Each condition gets 4 
   *parameters: 3 of them are of type iterator over a word of process symbols, 
   *and the last is of type symbol. 
   */
  using namespace process;
  using namespace predicate;
  using namespace transform;
  using namespace rtransform;

  typedef PWord::const_iterator Const_In;

  /**
   *class task.
   */
  class Task:public virtual Printable
  {
  public:
    /**
     * class reverse_generator.
     * this is just an interface.
     */ 
    class reverse_generator
    {
    public:
      typedef GExConst value_type;
      typedef Ref<reverse_generator> ref_type;
      
    public:
      virtual bool end() const=0;
      virtual reverse_generator& operator++()=0;
      virtual reverse_generator& operator=(const reverse_generator& other)=0;
      virtual value_type operator*() const=0;
      virtual ~reverse_generator() {}
    };

  public:
    typedef reverse_generator iterator_type;
    typedef reverse_generator::value_type value_type;
    
    virtual reverse_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const=0;
    virtual value_type operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const=0;
    virtual ~Task() {};
  };

  /**
   *class local task.
   */
  class Local_Task:public virtual Task
  {
  public:
    virtual ~Local_Task() {};
  };
  
  /**
   *class global task.
   */
  class Global_Task:public virtual Task
  {
  public:
    virtual ~Global_Task() {};
  };
  
  /**
   *class  TLocal_Task.
   */
  class TLocal_Task:public virtual Local_Task
  {
  public:
    typedef Task::reverse_generator::value_type value_type;

  public:
    TLocal_Task(const Transform<Process>* _t):t(_t) {}
      ~TLocal_Task(){delete(t);};
    value_type operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
    {
      Const_In nex_pos=pos; nex_pos++;PWord temp_w;
      temp_w.insert(temp_w.end(),nex_pos,w->end());
      temp_w.push_front(t->operator()(*pos));
      temp_w.insert(temp_w.begin(),w->begin(),pos);
      value_type result(temp_w,*gc);
      result.word.adjust(w->begin(),pos,w->end());
      return result;
    }

    void print(std::ostream& out) const {out << *t;}

    class reverse_generator:public virtual Task::reverse_generator
    {
    private:
      typedef RTransform<Process>::reverse_iterator_ref sub_type;

    public:
      reverse_generator() {}
      reverse_generator(const RTransform<Process>* _t,const PWord* _w, const Const_In _pos, const Global_Symbol* _gs):w(_w),pos(_pos),s(sub_type(&_t->begin(*pos))),gs(_gs) {}

      bool end() const
      {
	return (*s).end();
      }

      reverse_generator& operator++()
      {
	++(*s);
	return *this;
      }
      
      reverse_generator& operator=(const Task::reverse_generator& other)
	{
	  const reverse_generator* r=dynamic_cast<const reverse_generator*>(&other);
	  w=r->w;pos=r->pos;nex_pos=r->nex_pos;s=r->s;gs=r->gs;
	  return *this;
	}
      
      value_type operator*() const
      {
	Const_In nex_pos=pos; ++nex_pos;PWord temp_w;

	temp_w.insert(temp_w.end(),nex_pos,w->end());
	temp_w.push_front(**s);
	temp_w.insert(temp_w.begin(),w->begin(),pos);
	
	value_type result(temp_w,*gs);
	result.word.adjust(w->begin(),pos,w->end());
	
	return result;
      }

    private:
      const PWord* w;
      Const_In pos;
      Const_In nex_pos;
      sub_type s;
      const Global_Symbol* gs;
    };  

  public:
    reverse_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	const RTransform<Process>* r=dynamic_cast<const RTransform<Process>*>(t);
	assert(r);
	return *new reverse_generator(r,w,pos,gc);
      }

  private:
    const Transform<Process>* t;

  };
  
  /**
   *class  TGlobal_Task.
   */
  class TGlobal_Task:public virtual Global_Task
  {
  public:
    typedef Task::reverse_generator::value_type value_type;

  public:
    TGlobal_Task(const Transform<Global_Symbol>* _t):t(_t) {}

      value_type operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	value_type result(*w,(*t).operator()(*gc));
	result.word.adjust(w->begin(),pos,w->end());
	return result;
      }

    void print(std::ostream& out) const {out << *t;}

    class reverse_generator:public virtual Task::reverse_generator
    {
    private:
      typedef RTransform<Global_Symbol>::reverse_iterator_ref sub_type;
      
    public:
      reverse_generator() {}
      reverse_generator(const RTransform<Global_Symbol>* _t,const PWord* _w, const Const_In _pos, const Global_Symbol* _gs):w(_w),pos(_pos),s(sub_type(&_t->begin(*_gs))) {}

      bool end() const
      {
	return (*s).end();
      }

      reverse_generator& operator++()
      {
	++(*s);
	return *this;
      }
      
      reverse_generator& operator=(const Task::reverse_generator& other)
	{
	  const reverse_generator* r=dynamic_cast<const reverse_generator*>(&other);
	  w=r->w;pos=r->pos;s=r->s;
	  return *this;
	}
      
      value_type operator*() const
      {
	value_type result(*w,**s);
	result.word.adjust(w->begin(),pos,w->end());
	return result;
      }

    private:
      const PWord* w;
      Const_In pos;
      sub_type s;
    };  
public:
    reverse_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	const RTransform<Global_Symbol>* r=dynamic_cast<const RTransform<Global_Symbol>*>(t);
	assert(r);
	return *new reverse_generator(r,w,pos,gc);
      }

  private:
    const Transform<Global_Symbol>* t;
  };

  class TGlobal_Sequence_Task:public virtual Global_Task
  {
  public: 
    typedef Task::reverse_generator::value_type value_type;

  public:
    TGlobal_Sequence_Task(const Sequence_Transform<Process,Const_In>* _q):q(_q) {}
      
      value_type operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	value_type result;
	result.word=q->operator()(w->begin(), w->end(), pos);
	result.gs=(*gc);
	return result;
      }

      void print(std::ostream& out) const {out << *q;}
  
  public:
      class reverse_generator:public virtual Task::reverse_generator
      {
      private:
	typedef RSequence_Transform<Process,Const_In>::reverse_iterator_ref sub_type;
	
      public:
	reverse_generator() {}
	reverse_generator(const RSequence_Transform<Process,Const_In>* _t, const PWord* _w, const Const_In pos, const Global_Symbol* _gc):gs(_gc),s(sub_type(&_t->begin(_w->begin(),_w->end(),pos))) {}
	  
	  bool end() const { return (*s).end();}
	  
	  reverse_generator& operator++()
	    {
	      if (end()) {return *this;}
	      ++(*s);return *this;
	    }
	  
	  reverse_generator& operator=(const Task::reverse_generator& other)
	    {
	      const reverse_generator* r=dynamic_cast<const reverse_generator*>(&other);
	      assert(r);
	      gs=r->gs;s=r->s;
	      return *this;
	    }
	  
	  value_type operator*() const
	  {
	    value_type result;
	    result.word=**s;
	    result.gs=*gs;
	    return result;
	  }

      private:
	  const Global_Symbol* gs;
	  sub_type s;
      };

  public:
      reverse_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
	{
	  const RSequence_Transform<Process,Const_In>* r=dynamic_cast<const RSequence_Transform<Process,Const_In>*>(q);
	  assert(r);
	  return *new reverse_generator(r,w,pos,gc);
	}   

  private:
      const Sequence_Transform<Process,Const_In>* q;
  };

  /**
   *TGlobal_New_Task.
   * in forward should insert a new process in each place
   * in backward union of the possible one deletions.
   */
  /*  class TGlobal_New_Task:public virtual Global_Task
  {
  public:
    typedef Task::reverse_generator::value_type value_type;

  public:
    TGlobal_New_Task(const Transform<Global_Symbol>* _t):t(_t) {}

      value_type operator()(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	value_type result(*w,(*t).operator()(*gc));
	result.word.adjust(w->begin(),pos,w->end());
	return result;
      }

    void print(std::ostream& out) const {out << *t;}

    class reverse_generator:public virtual Task::reverse_generator
    {
    private:
      typedef RTransform<Global_Symbol>::reverse_iterator_ref sub_type;
      
    public:
      reverse_generator() {}
      reverse_generator(const RTransform<Global_Symbol>* _t,const PWord* _w, const Const_In _pos, const Global_Symbol* _gs):w(_w),pos(_pos),s(sub_type(&_t->begin(*_gs))) {}

      bool end() const
      {
	return (*s).end();
      }

      reverse_generator& operator++()
      {
	++(*s);
	return *this;
      }
      
      reverse_generator& operator=(const Task::reverse_generator& other)
	{
	  const reverse_generator* r=dynamic_cast<const reverse_generator*>(&other);
	  w=r->w;pos=r->pos;s=r->s;
	  return *this;
	}
      
      value_type operator*() const
      {
	value_type result(*w,**s);
	result.word.adjust(w->begin(),pos,w->end());
	return result;
      }

    private:
      const PWord* w;
      Const_In pos;
      sub_type s;
    };  
public:
    reverse_generator& begin(const PWord* w, const Const_In pos, const Global_Symbol* gc) const
      {
	const RTransform<Global_Symbol>* r=dynamic_cast<const RTransform<Global_Symbol>*>(t);
	assert(r);
	return *new reverse_generator(r,w,pos,gc);
      }

  private:
    const Transform<Global_Symbol>* t;
  };  
  */
}
  
#endif
  

