/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file transform.h
 *ADT for transforms.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */
#ifndef TRANSFORM_H
#define TRANSFORM_H

#include "symbol.h"
#include "tword.h"
#include "predicate.h"
#include "printable.h"
#include "exconst.h"
#include "ref.h"

#include <string>


//#define TRA_DEBUG
#ifdef TRA_DEBUG
#include <assert.h>
#endif

namespace transform
{
  using namespace predicate;
  using namespace starquantifier;

  /**
   *abstract class Transform over symbols.
   *it is parameterized by the type of output.
   *the tranform is legal only when T is a concret type of Symbol.
   */
  template<class T>
    class Transform:public virtual Printable
  {
  public:
    T operator()(const Symbol& s) const
    {
      const T* r=dynamic_cast<const T*>(&s);
#ifdef TRA_DEBUG
      assert(r);
#endif
      return operator()(*r); 
    }
    virtual T operator()(const T& s) const=0;
    virtual ~Transform() {}
  };
  
  /**
   *Obvious transform: the Identity.
   */
  template<class T>
    class Identity:public virtual Transform<T>
  {
  public:
    Identity() {}
    T operator()(const T& s) const
    {
      return T(s);
    }
    
    void print(std::ostream& out) const {out << "Id";}
  };
  
  /**
   *The assignment transform.
   *It assumes T of type map(id,val) where id is the name of variables and val as
   *possible assignment.
   */
  template<class T>
    class Assignment:public virtual Transform<T>
  {
  public:
    Assignment(const std::string _id, const int _val):id(_id),val(_val) { }
      
      T operator()(const T& s) const
      {
	T temp(s);
#ifdef TRA_DEBUG
	assert(s.get_definitions()->defined(id) && s.get_definitions()->in_range(id,val));
#endif
	temp[id]=val;
	return temp;
      }
      
      std::string get_id() const {return id;}
      int get_val() const {return val;}
      void print(std::ostream& out) const {out << "p." << id << ":=" << val;}
  
  private:
      const std::string id;
      const int val;
  }; 
  
  /**
   *Abstract class transform over sequence.
   *return type is always of the form TWord.
   *parameter T for TWord template parameter.
   */
  template<class T,class Seq>
    class Sequence_Transform:public virtual Printable
     {
     public:
       typedef ExConst<T> value_type;
     public:
       virtual value_type operator()(const Seq first, const Seq last, const Seq curr) const=0;
       virtual ~Sequence_Transform() {}
     };
   
  /**
   *Natural Sequence_Transform, transforms everything except the current.
   */
  template<class T,class Seq>
    class Sequence_Transform_A:public virtual Sequence_Transform<T,Seq>
  {
  public:
    typedef typename Sequence_Transform<T,Seq>::value_type value_type;

  public:
    Sequence_Transform_A(const Transform<T>* _t):t(_t) {}
      value_type operator()(const Seq first, const Seq last, const Seq curr) const
	{
	  TWord<T> w;
	  Seq iter=first;
	  for( ; iter!=last ; iter++)
	    {
	      if (iter==curr) {w.push_back(*curr);}
	      else {w.push_back(t->operator()(*iter));}
	    }
	  value_type result(w);
	  result.adjust(first,curr,last);
	  return result;
	}
      
      const Transform<T>* get_trans() const {return t;}
      
      void print(std::ostream& out) const
      {
	out << "Forall p except the current : " << *t;
      };
      
  private:
      const Transform<T>* t;
  };

  /**
   *Natural Sequence transform: transform if a certain predicate holds.
   */
  template<class T,class Seq>
    class Sequence_Transform_If
  {
    typedef typename Sequence_Transform<T,Seq>::value_type value_type;

  public:
    Sequence_Transform_If(const Predicate* _p,const Transform<T>* _t):p(_p),t(_t) {}
      value_type operator()(const Seq first, const Seq last, const Seq curr) const
	{
	  TWord<T> w;
	  Seq iter=first;
	  for( ; iter!=last ; iter++)
	    {
	      if (p->operator()(*iter)) {w.push_back(t->operator()(*iter));}
	      else {w.push_back(*iter);}
	    }
	  value_type result(w);
	  result.adjust(first,curr,last);
	  return result;
	}
      
      const Predicate* get_pred() const {return p;}
      const Transform<T>* get_trans() const {return t;}
  
  private:
      const Predicate* p;
      const Transform<T>* t;
  };
   
  /**
   *Derived sequence transform from sequence transform if.
   *It transforms only to the left of a given current position.
   */
  template<class T,class Seq>
    class Sequence_Transform_IfL:private Sequence_Transform_If<T,Seq>,public virtual Sequence_Transform<T,Seq>
    {
    private:
      typedef Sequence_Transform_If<T,Seq> base_type;
      
    public:
      typedef typename Sequence_Transform<T,Seq>::value_type value_type;

    public:
      Sequence_Transform_IfL(const Predicate* _p,const Transform<T>* _t):base_type(_p,_t) {}
	
      value_type operator()(const Seq first, const Seq last, const Seq curr) const
	{
	  Ref<Predicate> p(new True());
	  Ref<Transform<T> > t(new Identity<T>());
	  Sequence_Transform_If<T,Seq> st1(&(*p),&(*t));
	  TWord<T> temp((base_type::operator()(first,curr,curr)).word);
	  temp.merge(st1(curr,last,curr).word);
	  value_type result(temp);
	  result.adjust(first,curr,last);
	  return result;
	}
      
      const Predicate* get_pred() const {return base_type::get_pred();}
      const Transform<T>* get_trans() const {return base_type::get_trans();}
      
      void print(std::ostream& out) const {out << "ForAll p to the left, if " << *get_pred() << " then " << *get_trans();}
    };
  
  /**
   *Derived sequence transform from sequence transform if.
   *It transforms only to the right of a given current position.
   */
  template<class T,class Seq>
    class Sequence_Transform_IfR:private Sequence_Transform_If<T,Seq>,public virtual Sequence_Transform<T,Seq>
    {
    private:
      typedef Sequence_Transform_If<T,Seq> base_type;  

    public:
      typedef typename Sequence_Transform<T,Seq>::value_type value_type;
      
    public:
      Sequence_Transform_IfR(const Predicate* _p,const Transform<T>* _t):base_type(_p,_t) {}
	
	value_type operator()(const Seq first, const Seq last, const Seq curr) const
	  {
	    Ref<Predicate> p(new True());
	    Ref<Transform<T> > t(new Identity<T>);
	    Sequence_Transform_If<T,Seq> st1(&(*p),&(*t));
	    if (curr!=last) 
	      {
		Seq iter=curr;iter++;
		TWord<T> temp(st1(first,iter,last).word);
		temp.merge((base_type::operator()(iter,last,curr)).word);
		value_type result(temp);
		result.adjust(first,curr,last);
		return result;
	      }
	    else {return st1(first,last,curr);}
	  }
	
	const Predicate* get_pred() const {return base_type::get_pred();}
	const Transform<T>* get_trans() const {return base_type::get_trans();}
	
	void print(std::ostream& out) const {out << "ForAll p to the right, if " << *get_pred() << " then " << *get_trans();}
    };

}

#endif

