/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file tword.h
 *ADT for list-based words in general (parameterized by the type of the symbols).
 *Besides it implements order and upward closed sets of words with respect to that
 *order.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef TWORD_H
#define TWORD_H
#include <cstdlib>
#include <list>
#include<iostream>

#include "word.h"
#include "minset.h"
#include "ref.h"

/**
 *This class implements words based on lists. 
 */
template<class T>
class TWord:private std::list<T>, public Word
{
 public:
 typedef std::list<T> ContainerT;

 public:
  TWord() {}

 public:
  typedef typename ContainerT::const_iterator const_iterator;
  typedef typename ContainerT::iterator iterator;
  typedef typename ContainerT::const_reverse_iterator const_reverse_iterator;
  typedef typename ContainerT::reverse_iterator reverse_iterator;

 public:

  /**emptyness check*/
  unsigned int size() const { return ContainerT::size(); }
  bool empty() const { return ContainerT::empty(); }

  /**iterators*/
  iterator begin() { return ContainerT::begin(); }
  const_iterator begin() const { return ContainerT::begin(); }

  iterator end() { return ContainerT::end(); }
  const_iterator end() const { return ContainerT::end(); }

  reverse_iterator rbegin() { return ContainerT::rbegin(); }
  const_reverse_iterator rbegin() const { return ContainerT::rbegin(); }
    
  reverse_iterator rend() { return ContainerT::rend(); }
  const_reverse_iterator rend() const { return ContainerT::rend(); }
  
  /**operations*/  
  void swap(TWord<T> other) { ContainerT::swap(other); }
  void clear() { ContainerT::clear(); }

  void erase(iterator it) { ContainerT::erase(it); }
  void erase(const_iterator it) { ContainerT::erase(it); }

  /**addition*/
  void push_back(const T& x) { ContainerT::push_back(x); }
  void push_front(const T& x) { ContainerT::push_front(x); }
  void insert(iterator pos, const T& x) { ContainerT::insert(pos,x); }
  void insert(iterator pos, const_iterator f, const_iterator l)
  {
    const_iterator t=f;
    while(t!=l)
      {
	ContainerT::insert(pos,*t);
	++t;
      }
  }

  void merge(const TWord<T>& w){ insert(end(),w.begin(),w.end()); }

  /**inherited from Word*/
  bool less(const Word& W) const
  {
    const TWord<T>* PW=dynamic_cast<const TWord<T>* >(&W);
    assert(PW);
    return less(*PW);
  }
 
  /**inherited from Printable via Word*/
  void print(std::ostream& out) const
  {
    const_iterator p = begin();
    
    //    for (; p != end(); ++p) out << "\t" << *p << "\n";
    for (; p != end(); ++p) out << *p ;
  }

 public:
 
  bool less(const TWord<T>& other) const
  {
    const_iterator p=begin(),p_end=end(),q=other.begin(),q_end=other.end();
    while(p != p_end)
      {	
	if (q == q_end) {return false;}
	if (*p == *q) {++p;++q;}
	else {++q;}
      }

    return true;
  }

};

/**
 *This class implements order on words
 */
template<class T>
class TWordOrder
{
 public:
  TWordOrder() {}

 bool operator()(const TWord<T>& W1, const TWord<T>& W2) const
  {
    return W1 <= W2;
  }
 bool operator()(const Ref<const TWord<T> >& W1, const Ref<const TWord<T> >& W2) const
  {
    return *W1 <= *W2;
  }

};


/**
 *Tword equality.
 */
template<class T>
class TWordEquality
{
 public:
  TWordEquality() {}

 bool operator()(const TWord<T>& W1, const TWord<T>& W2) const
  {
    return W1 <= W2 && W2 <= W1;
  }
 bool operator()(const Ref<const TWord<T> >& W1, const Ref<const TWord<T> >& W2) const
  {
    return *W1 <= *W2 && *W2 <= *W1;
  }

};

/**
 *This class implements upward closed set of words.
 */
template<class T>
class UCSetTWord:public MinSet<Ref<const TWord<T> > , TWordOrder<T> >
{

 public:
  UCSetTWord() {}

};

/**
 *Set fo twords.
 */
template<class T>
class SetTWord:public MinSet<Ref<const TWord<T> > , TWordEquality<T> >
{

 public:
  SetTWord() {}

};
#endif
