/*
    Copyright (C) 2005 Ahmed Rezine and Noomene Ben Henda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information: Ahmed Rezine <Rezine.Ahemd@it.uu.se> and Noomene Ben Henda <Noomene.BenHenda@it.uu.se>
*/

/**@file variable.h
 *ADT for variables.
 *
 *@author Ahmed Rezine and Noomene Ben-Henda
 */

#ifndef VARIABLE_H
#define VARIABLE_H

#include "symbol.h"
#include "process.h"
#include "starpredicate.h"
#include "predicate.h"

//#define VAR_DEBUG
#ifdef VAR_DEBUG
#include <assert.h>
#endif

namespace variable
{
  using namespace process;
  using namespace predicate;
  using namespace starpredicate;
  /**
   *class Variable
   *it is parameterized by a given domain or object type.
   *T is of type map(id,val) where id is the name of the variable.
   */
  template<class T>
    class Variable:public virtual Function
  {
  public:
    Variable(Var_Id _id):id(_id) {}

      bool holds_for_subset(const Symbol& s) const
      {
	const T* r=dynamic_cast<const T*>(&s);
#ifdef VAR_DEBUG
	assert(r);
#endif
	return holds_for_subset(*r);
      } 

      int operator()(const Symbol& s) const
      {
	const T* r=dynamic_cast<const T*>(&s);
#ifdef VAR_DEBUG
	assert(r);
#endif
	return operator()(*r);
      }

      bool holds_for_subset(const T& s) const
      {
	return (s.get_definition()->star_enabled(id)) && (s.find(id)->second==s.get_definition()->star_value(id));
      }

      int operator()(const T& s) const
      {
	return s.find(id)->second;
      }

      Var_Id get_id() const {return id;}
      
      void print(std::ostream& out) const {out << "p." << id;}

  private:
      Var_Id id;

  };

  template<class T>
    class StarVariable:public virtual StarFunction<T>,public Variable<T>
  { 
  private:
    typedef Variable<T> base_type;

  public:
    StarVariable(Var_Id _id):base_type(_id) {}

      class constraint_iterator:public StarFunction<T>::constraint_iterator
      {
	
      public:
	constraint_iterator() {}
	constraint_iterator(const StarVariable<T>* _v,const T* _s,bool _e):v(_v),s(*_s),e(_e)
	  {
	    if (!(_v->holds_for_subset(s))){i=0;e=true;}
	    else 
	      {
		i=s.get_definition()->lower_bound(v->get_id());
		s[v->get_id()]=i;
	      }
	  }
	  
	  bool end() const { return e;}
	  
	  constraint_iterator& operator++()
	    {
	      if (e) {return *this;}
	      ++i;
	      if (i>=s.get_definition()->star_value(v->get_id()))
		{e=true;return *this;}
	      else
		{s[v->get_id()]=i;return *this;}
	    }
	  
	  constraint_iterator& operator=(const typename StarFunction<T>::constraint_iterator& other)
	    {
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      v=r->v;s=r->s;e=r->e;i=r->i;
	      return *this;
	    }
	  
	  T operator*() const
	  {
	    return s;
	  }
	  
      protected:
	  const StarVariable<T>* v;
	  T s;
	  bool e;
	  int i;
      };

  public:
      constraint_iterator& begin(const T& s) const
	{
	  return *new constraint_iterator(this,&s,false);
	}	

      StarFunction<T>& negate() const
	{
	  return *new StarVariable(base_type::get_id());
	}

  };

  /**
   *class Constant
   *it is universal and doesn't need a parameter. However constants here range over 
   *integers
   */
  class Constant:public virtual Function
  {
  public:
    Constant(int _val):val(_val) {}

      bool holds_for_subset(const Symbol& s) const {return false;}
      
      int operator()(const Symbol& s) const {return val;}
      
      int get_val() const {return val;}

      void print(std::ostream& out) const {out << val;}

  private:
      int val;
  };


  /**
   *class StarConstant
   *The parameter is just for the beauty.
   */
  template<class T>
    class StarConstant:public virtual StarFunction<T>,public Constant
  {
  private:
    typedef Constant base_type;

  public:
    StarConstant(int _val):base_type(_val) {}

      class constraint_iterator:public StarFunction<T>::constraint_iterator
      {
	
      public:
	constraint_iterator() {}
	constraint_iterator(const T* _s):s(*_s) {e=true;}
	  
	  bool end() const { return e;}
	  
	  constraint_iterator& operator++() { return *this; }
	  
	  constraint_iterator& operator=(const typename StarFunction<T>::constraint_iterator& other)
	    {
	      const constraint_iterator* r=dynamic_cast<const constraint_iterator*>(&other);
	      s=r->s;e=r->e; return *this;
	    }
	  
	  T operator*() const
	  {
	    return s;
	  }
	  
      protected:
	  T s;
	  bool e;
      };

  public:
      constraint_iterator& begin(const T& s) const
	{
	  return *new constraint_iterator(&s);
	}	
      
      StarFunction<T>& negate() const 
	{
	  return *new StarConstant(base_type::get_val());
	}
  };

}

#endif

